import warnings
from typing import Union, Set

import pytest
from pybedtools import BedTool

warnings.filterwarnings("ignore", category=DeprecationWarning)

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment

from genomics_data_index.test.integration import data_dir, masked_positions_snippy_bed
from genomics_data_index.storage.service.CoreAlignmentService import CoreAlignmentService


def remove_column(alignment: MultipleSeqAlignment, position: int) -> MultipleSeqAlignment:
    if position <= 0 or position > alignment.get_alignment_length():
        raise Exception(f'position [{position}] is in an invalid range')

    return alignment[:, :(position - 1)] + alignment[:, position:]


def replace_column_with_reference(alignment: MultipleSeqAlignment, position: Union[int, range],
                                  skip_missing=True) -> MultipleSeqAlignment:
    if isinstance(position, range):
        for pos in position:
            alignment = replace_column_with_reference_internal(alignment, position=pos, skip_missing=skip_missing)
        return alignment
    else:
        return replace_column_with_reference_internal(alignment, position=position, skip_missing=skip_missing)


def replace_column_with_reference_internal(alignment: MultipleSeqAlignment, position: int,
                                           skip_missing=True) -> MultipleSeqAlignment:
    column = alignment[:, (position - 1):position]
    ref = None
    ref_name = 'genome'
    for s in column:
        if s.id == ref_name:
            ref = s

    if ref is None:
        raise Exception('Could not find reference')

    for s in column:
        if s.id != ref_name:
            ss = s.seq.upper()
            if (not skip_missing) or (ss != 'N' and ss != '-'):
                s.seq = ref.seq

    return alignment[:, :(position - 1)] + column + alignment[:, position:]


def replace_ref_name(alignment):
    # Change reference name
    for s in alignment:
        if s.id == 'Reference':
            s.id = 'genome'
            s.description = '[reference genome]'
        else:
            s.description = 'generated automatically'


def replace_gap_with_n_and_upper(alignment: MultipleSeqAlignment,
                                 position_set: Set[int] = None) -> MultipleSeqAlignment:
    for record in alignment:
        seq = record.seq.tomutable()
        for index, character in enumerate(seq):
            if position_set is None or (index + 1) in position_set:
                if character.upper() == '-' or character.upper() == 'X':
                    seq[index] = 'N'
                else:
                    seq[index] = seq[index].upper()

        record.seq = seq.toseq()
    return alignment


def replace_n_with_gap_and_upper(alignment: MultipleSeqAlignment,
                                 position_set: Set[int] = None) -> MultipleSeqAlignment:
    for record in alignment:
        seq = record.seq.tomutable()
        for index, character in enumerate(seq):
            if position_set is None or (index + 1) in position_set:
                if character.upper() == 'N':
                    seq[index] = '-'
                else:
                    seq[index] = seq[index].upper()

        record.seq = seq.toseq()
    return alignment


def alignment_upper(alignment: MultipleSeqAlignment):
    for record in alignment:
        record.seq = record.seq.upper()
    return alignment


@pytest.fixture
def expected_alignment_core() -> MultipleSeqAlignment:
    '''
    Loads the expected alignment for testing (core alignment).
    This expected alignment was copied from the output of snippy. However, snippy
    includes some complex events in the alignment which I do not include.
    I modify this expected alignment (from snippy) by subtracting 2 SNVs/SNPs introduced
    by complex events. That way I can compare the alignments for equality.
    :return: The expected alignment as a MultipleSeqAlignment object.
    '''
    with open(data_dir / 'snippy.core.aln', 'r') as f:
        alignments = list(AlignIO.parse(f, 'fasta'))
        alignment = alignments[0]

        replace_ref_name(alignment)

        # Remove SNVs/SNPs introduced by complex events
        alignment = remove_column(alignment, 15)  # Position 1325
        alignment = remove_column(alignment, 23 - 1)  # Position 1984 (-1 to account for previous removed column)

        alignment.sort()

        return alignment


@pytest.fixture
def expected_alignment_full_unmodified() -> MultipleSeqAlignment:
    """
    Loads the expected alignment for testing (full alignment).
    This expected alignment was copied from the output of snippy.
    :return: The exepcted alignment as a MultipleSeqAlignment object.
    """
    with open(data_dir / 'snippy.full.aln', 'r') as f:
        alignments = list(AlignIO.parse(f, 'fasta'))
        alignment = alignments[0]

        replace_ref_name(alignment)
    return alignment


@pytest.fixture
def expected_alignment_full_only_snps(expected_alignment_full_unmodified) -> MultipleSeqAlignment:
    """
    Loads the full snippy alignment, and then removes all but SNPs.
    :return: The exepcted alignment as a MultipleSeqAlignment object.
    """
    alignment = expected_alignment_full_unmodified

    # Replace deletions/complex
    # Generated using
    # zgrep -hv '^#' Sample*/snps.vcf.gz | zgrep 'del\|complex' | cut -f 2,4 | sort -nu |
    #  sed -e 's/^\([0-9][0-9]*\)\t\(\w*\)/alignment =
    #   replace_column_with_reference(alignment, range(\1 + 1, \1 + 1 + len("\2") - 1), skip_missing=False)/'
    alignment = replace_column_with_reference(alignment, range(302 + 1, 302 + 1 + len("CT") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(347 + 1, 347 + 1 + len("TGAAG") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(349 + 1, 349 + 1 + len("AAGT") - 1),
                                              skip_missing=False)

    # This is a complex/OTHER event, which won't be represented in my alignment
    # Needed to expand boundaries on the left/right by 1 since it's not a simple deletion
    alignment = replace_column_with_reference(alignment, range(461, 461 + len("AAAT")),
                                              skip_missing=False)

    alignment = replace_column_with_reference(alignment, range(866 + 1, 866 + 1 + len("GCCAGATCC") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(883 + 1, 883 + 1 + len("CACATG") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1070 + 1, 1070 + 1 + len("TG") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1135 + 1, 1135 + 1 + len("CCT") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1167 + 1, 1167 + 1 + len("GA") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1209 + 1, 1209 + 1 + len("AG") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1270 + 1, 1270 + 1 + len("CTT") - 1),
                                              skip_missing=False)

    # This is a complex/OTHER event, so won't be represented in my alignment. Need to treat left/right
    # boundaries differently
    alignment = replace_column_with_reference(alignment, range(1325, 1325 + len("TGA")),
                                              skip_missing=False)

    alignment = replace_column_with_reference(alignment,
                                              range(1483 + 1, 1483 + 1 + len("AAAGAGGGGCTGCTGGAGCCG") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1708 + 1, 1708 + 1 + len("ATGCTGTTCAATAC") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(1815 + 1, 1815 + 1 + len("CAA") - 1),
                                              skip_missing=False)

    # This is a complex/OTHER event, so won't be represented in my alignment. Need to treat left/right
    # boundaries differently
    alignment = replace_column_with_reference(alignment, range(1984, 1984 + len("GTGATTG")),
                                              skip_missing=False)

    alignment = replace_column_with_reference(alignment, range(2419 + 1, 2419 + 1 + len("GA") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(2480 + 1, 2480 + 1 + len("TG") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(3008 + 1, 3008 + 1 + len("GA") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(3276 + 1, 3276 + 1 + len("GC") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(3576 + 1, 3576 + 1 + len("TTTC") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(3656 + 1, 3656 + 1 + len("CATT") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(3897 + 1, 3897 + 1 + len("GCGCA") - 1),
                                              skip_missing=False)
    alignment = replace_column_with_reference(alignment, range(4437 + 1, 4437 + 1 + len("AG") - 1),
                                              skip_missing=False)

    alignment = replace_gap_with_n_and_upper(alignment)

    alignment.sort()

    return alignment


@pytest.fixture
def expected_alignment_full_snps_mnp_deletions(expected_alignment_full_unmodified) -> MultipleSeqAlignment:
    """
    Loads the full snippy alignment, and then keeps SNPs, MNPs (which aren't in test data) and DELETIONs
    :return: The exepcted alignment as a MultipleSeqAlignment object.
    """
    alignment = alignment_upper(expected_alignment_full_unmodified)

    masked_positions = BedTool(masked_positions_snippy_bed)
    positions_to_fill_with_n = set()
    for interval in masked_positions:
        positions_to_fill_with_n = positions_to_fill_with_n | set(range(interval.start + 1, interval.end + 1))

    alignment = replace_gap_with_n_and_upper(alignment, position_set=positions_to_fill_with_n)

    # This is a complex/OTHER event, which won't be represented in my alignment
    # Needed to expand boundaries on the left/right by 1 since it's not a simple deletion
    alignment = replace_column_with_reference(alignment, range(461, 461 + len("AAAT")),
                                              skip_missing=False)

    # This is a complex/OTHER event, so won't be represented in my alignment. Need to treat left/right
    # boundaries differently
    alignment = replace_column_with_reference(alignment, range(1325, 1325 + len("TGA")),
                                              skip_missing=False)

    # This is a complex/OTHER event, so won't be represented in my alignment. Need to treat left/right
    # boundaries differently
    alignment = replace_column_with_reference(alignment, range(1984, 1984 + len("GTGATTG")),
                                              skip_missing=False)

    # These are recorded as deletions in the VCF file, but snippy is setting some characters to N instead of -
    # Since I set them as - in my results I need to change them in the expected alignment
    alignment = replace_n_with_gap_and_upper(alignment, {1210, 2420, 3657, 3658, 3659, 4438,
                                                         885, 1272, 2481,
                                                         1136, 1816, 3009})

    alignment.sort()

    return alignment


def get_unequal_positions_str(expected, actual) -> str:
    return_val = ''
    for idx, e in enumerate(expected.seq):
        a = actual[idx]
        if e != a:
            return_val = return_val + f'{idx + 1}: expected:{expected.id}={e} != actual:{actual.id}={a}\n'

    return return_val


def compare_alignments(expected: MultipleSeqAlignment, actual: MultipleSeqAlignment) -> None:
    assert len(expected) == len(actual), 'Alignment has incorrect number of samples'
    assert expected.get_alignment_length() == actual.get_alignment_length()

    number_of_sequences = len(expected)
    for i in range(0, number_of_sequences):
        assert expected[i].id == actual[i].id, f'Alignment ids are not equal [{expected[i].id}] != [{actual[i].id}]'
        assert expected[i].seq == actual[i].seq, f'Alignment sequences are not equal for ' \
                                                 f'[expected.id={expected[i].id}, actual.id={actual[i].id}] ' \
                                                 f'\n{get_unequal_positions_str(expected[i], actual[i])}'


def test_snippy_core_align(core_alignment_service: CoreAlignmentService, expected_alignment_core):
    actual_alignment = core_alignment_service.construct_alignment(reference_name='genome',
                                                                  samples=['SampleA', 'SampleB', 'SampleC'],
                                                                  include_variants=['SNP'])
    compare_alignments(expected_alignment_core, actual_alignment)


def test_snippy_core_align_with_other_include(core_alignment_service: CoreAlignmentService):
    with pytest.raises(Exception) as execinfo:
        core_alignment_service.construct_alignment(reference_name='genome',
                                                   samples=['SampleA', 'SampleB', 'SampleC'],
                                                   include_variants=['MNP'])
    assert 'Currently align_type=core only works with include_variants=["SNP"]' in str(execinfo.value)


def test_snippy_full_align_with_invalid_include_variants(core_alignment_service: CoreAlignmentService):
    with pytest.raises(Exception) as execinfo:
        core_alignment_service.construct_alignment(reference_name='genome',
                                                   samples=['SampleA', 'SampleB', 'SampleC'],
                                                   align_type='full',
                                                   include_variants=['invalid'])
    assert f"Only {['SNP', 'MNP', 'DELETION', 'DELETION_OTHER']} are supported" in str(execinfo.value)


def test_snippy_full_align(core_alignment_service, expected_alignment_full_only_snps):
    actual_alignment = core_alignment_service.construct_alignment(reference_name='genome',
                                                                  samples=['SampleA', 'SampleB', 'SampleC'],
                                                                  align_type='full',
                                                                  include_variants=['SNP'])
    compare_alignments(expected_alignment_full_only_snps, actual_alignment)


def test_snippy_full_align_snp_mnp_deletion(core_alignment_service, expected_alignment_full_snps_mnp_deletions):
    actual_alignment = core_alignment_service.construct_alignment(reference_name='genome',
                                                                  samples=['SampleA', 'SampleB', 'SampleC'],
                                                                  align_type='full')
    compare_alignments(expected_alignment_full_snps_mnp_deletions, actual_alignment)


def test_regular_vcf_full_align(core_alignment_service_non_snippy_vcfs, expected_alignment_full_only_snps):
    actual_alignment = core_alignment_service_non_snippy_vcfs.construct_alignment(reference_name='genome',
                                                                                  samples=['SampleA', 'SampleB',
                                                                                           'SampleC'],
                                                                                  align_type='full',
                                                                                  include_variants=['SNP'])
    compare_alignments(expected_alignment_full_only_snps, actual_alignment)


def test_regular_vcf_core_align(core_alignment_service_non_snippy_vcfs, expected_alignment_core):
    actual_alignment = core_alignment_service_non_snippy_vcfs.construct_alignment(reference_name='genome',
                                                                                  samples=['SampleA', 'SampleB',
                                                                                           'SampleC'],
                                                                                  include_variants=['SNP'])
    compare_alignments(expected_alignment_core, actual_alignment)



import tempfile
from os import path
from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from pybedtools import BedTool

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions

root_data_dir = Path(path.dirname(__file__), 'data')
snps_vcf_mask_file = root_data_dir / 'snps-for-mask.vcf.gz'


def test_create_from_sequence():
    sequences = [SeqRecord(seq=Seq('ATCG-NN'), id='record1')]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)

    assert 3 == len(mask), 'Invalid length'
    assert not mask.is_empty()
    assert not mask.contains('record1', 0)
    assert not mask.contains('record1', 3)
    assert not mask.contains('record1', 4, start_position_index='1')
    assert mask.contains('record1', 4)
    assert mask.contains('record1', 5, start_position_index='1')
    assert mask.contains('record1', 6)
    assert not mask.contains('record1', 7)


def test_mask_to_features_from_sequence():
    sequences = [SeqRecord(seq=Seq('ATCG-NN'), id='record1')]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)
    features_df = mask.mask_to_features().sort_values('POS')

    assert ['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'] == features_df.columns.tolist()
    assert 3 == len(features_df)
    assert ['record1:5:1:?', 'record1:6:1:?', 'record1:7:1:?'] == features_df['VARIANT_ID'].tolist()


def test_mask_to_features_from_sequence_no_mask():
    sequences = [SeqRecord(seq=Seq('ATCGAAA'), id='record1')]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)
    features_df = mask.mask_to_features().sort_values('POS')

    assert ['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'] == features_df.columns.tolist()
    assert 0 == len(features_df)


def test_mask_to_features_from_bed():
    mask = MaskedGenomicRegions(BedTool('reference 0 4', from_string=True))
    features_df = mask.mask_to_features().sort_values('POS')

    assert ['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'] == features_df.columns.tolist()
    assert 4 == len(features_df)
    assert ['reference:1:1:?', 'reference:2:1:?', 'reference:3:1:?',
            'reference:4:1:?'] == features_df['VARIANT_ID'].tolist()


def test_create_from_two_sequences():
    sequences = [
        SeqRecord(seq=Seq('ATCG-NN'), id='record1'),
        SeqRecord(seq=Seq('NN-GAT'), id='record2')
    ]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)

    assert 6 == len(mask), 'Invalid length'
    assert not mask.is_empty()
    assert not mask.contains('record1', 0)
    assert not mask.contains('record1', 3)
    assert mask.contains('record1', 4)
    assert mask.contains('record1', 6)
    assert not mask.contains('record1', 7)

    assert mask.contains('record2', 0)
    assert mask.contains('record2', 2)
    assert not mask.contains('record2', 3)
    assert not mask.contains('record2', 5)


def test_create_from_vcf():
    mask = MaskedGenomicRegions.from_vcf_file(snps_vcf_mask_file)

    assert 9 == len(mask), 'Invalid length'
    assert not mask.is_empty()

    assert not mask.contains('reference', 291)
    assert mask.contains('reference', 292)
    assert not mask.contains('reference', 293)

    assert not mask.contains('reference', 300)
    assert mask.contains('reference', 301)
    assert mask.contains('reference', 302)
    assert not mask.contains('reference', 303)

    assert not mask.contains('reference', 372)
    assert mask.contains('reference', 373)
    assert not mask.contains('reference', 374)

    assert not mask.contains('reference', 459)
    assert mask.contains('reference', 460)
    assert mask.contains('reference', 463)
    assert not mask.contains('reference', 464)

    assert not mask.contains('reference', 504)
    assert mask.contains('reference', 505)
    assert not mask.contains('reference', 506)


def test_mask_to_features_from_two_sequences():
    sequences = [
        SeqRecord(seq=Seq('ATCG-NN'), id='record1'),
        SeqRecord(seq=Seq('NN-GAT'), id='record2')
    ]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)
    features_df = mask.mask_to_features().sort_values(['CHROM', 'POS'])

    assert ['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'] == features_df.columns.tolist()
    assert 6 == len(features_df)
    assert ['record1:5:1:?', 'record1:6:1:?', 'record1:7:1:?',
            'record2:1:1:?', 'record2:2:1:?', 'record2:3:1:?'] == features_df['VARIANT_ID'].tolist()


def test_mask_to_features_multi_range():
    mask = MaskedGenomicRegions(BedTool([('ref', 10, 15), ('ref', 20, 22)]))
    expected_number_features = 7

    features_df = mask.mask_to_features()
    features_df = features_df.sort_values('POS')

    assert ['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'] == features_df.columns.tolist()
    assert expected_number_features == len(features_df)
    assert ['ref'] * expected_number_features == features_df['CHROM'].tolist()
    assert [11, 12, 13, 14, 15, 21, 22] == features_df['POS'].tolist()
    assert [1] * expected_number_features == features_df['REF'].tolist()
    assert ['?'] * expected_number_features == features_df['ALT'].tolist()
    assert ['UNKNOWN_MISSING'] * expected_number_features == features_df['TYPE'].tolist()
    assert [f'ref:{p}:1:?' for p in [11, 12, 13, 14, 15, 21, 22]] == features_df['VARIANT_ID'].tolist()


def test_mask_to_features_multi_sequence():
    mask = MaskedGenomicRegions(BedTool([('ref', 10, 15), ('ref2', 12, 17)]))
    expected_nf = 5
    expected_nf2 = 5
    expected_t = 10

    features_df = mask.mask_to_features()
    features_df = features_df.sort_values(['CHROM', 'POS'])

    assert ['CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'VARIANT_ID'] == features_df.columns.tolist()
    assert expected_t == len(features_df)
    assert ['ref'] * expected_nf + ['ref2'] * expected_nf2 == features_df['CHROM'].tolist()
    assert [11, 12, 13, 14, 15, 13, 14, 15, 16, 17] == features_df['POS'].tolist()
    assert [1] * expected_t == features_df['REF'].tolist()
    assert ['?'] * expected_t == features_df['ALT'].tolist()
    assert ['UNKNOWN_MISSING'] * expected_t == features_df['TYPE'].tolist()
    assert [f'ref:{p}:1:?' for p in [11, 12, 13, 14, 15]] + [
        f'ref2:{p}:1:?' for p in [13, 14, 15, 16, 17]] == features_df['VARIANT_ID'].tolist()


def test_sequence_names():
    sequences = [
        SeqRecord(seq=Seq('ATCG-NN'), id='record1'),
        SeqRecord(seq=Seq('NN-GAT'), id='record2')
    ]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)
    assert {'record1', 'record2'} == mask.sequence_names()


def test_create_from_sequence_all_masked():
    sequences = [SeqRecord(seq=Seq('---------'), id='record1')]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)

    assert 9 == len(mask), 'Invalid length'
    assert mask.contains('record1', 0)
    assert mask.contains('record1', 8)


def test_create_empty():
    mask = MaskedGenomicRegions.empty_mask()

    assert 0 == len(mask), 'Invalid length'
    assert mask.is_empty()
    assert not mask.contains('record1', 0)
    assert not mask.contains('record1', 3)


def test_intersection():
    sequences1 = [SeqRecord(seq=Seq('ATCG-NN'), id='record1')]
    sequences2 = [SeqRecord(seq=Seq('ATCGNNTCC'), id='record1')]
    mask1 = MaskedGenomicRegions.from_sequences(sequences=sequences1)
    mask2 = MaskedGenomicRegions.from_sequences(sequences=sequences2)

    mask_intersect = mask1.intersect(mask2)

    assert 2 == len(mask_intersect), 'Invalid length'
    assert not mask_intersect.contains('record1', 3)
    assert mask_intersect.contains('record1', 4)
    assert mask_intersect.contains('record1', 5)
    assert not mask_intersect.contains('record1', 6)


def test_union_all():
    sequences1 = [SeqRecord(seq=Seq('ATCG-NN'), id='record1')]
    sequences2 = [SeqRecord(seq=Seq('ATCGNNTCC'), id='record1')]
    mask1 = MaskedGenomicRegions.from_sequences(sequences=sequences1)
    mask2 = MaskedGenomicRegions.from_sequences(sequences=sequences2)

    mask_union = MaskedGenomicRegions.union_all([mask1, mask2])

    assert 3 == len(mask_union), 'Invalid length'
    assert not mask_union.contains('record1', 3)
    assert mask_union.contains('record1', 4)
    assert mask_union.contains('record1', 5)
    assert mask_union.contains('record1', 6)
    assert not mask_union.contains('record1', 7)


def test_union():
    sequences1 = [SeqRecord(seq=Seq('ATCG-NN'), id='record1')]
    sequences2 = [SeqRecord(seq=Seq('ATCGNNTCC'), id='record1')]
    mask1 = MaskedGenomicRegions.from_sequences(sequences=sequences1)
    mask2 = MaskedGenomicRegions.from_sequences(sequences=sequences2)

    mask_union = mask1.union(mask2)

    assert 3 == len(mask_union), 'Invalid length'
    assert not mask_union.contains('record1', 3)
    assert mask_union.contains('record1', 4)
    assert mask_union.contains('record1', 5)
    assert mask_union.contains('record1', 6)
    assert not mask_union.contains('record1', 7)


def test_subtract():
    mask1 = MaskedGenomicRegions(BedTool([('ref', 10, 15)]))
    mask2 = MaskedGenomicRegions(BedTool([('ref', 12, 20)]))
    mask3 = MaskedGenomicRegions(BedTool([('ref', 10, 11), ('ref', 14, 15)]))
    mask4 = MaskedGenomicRegions(BedTool([('ref', 10, 11), ('ref2', 14, 15)]))
    mask_empty = MaskedGenomicRegions.empty_mask()

    mask_subtract = mask1.subtract(mask2)
    assert not mask_subtract.contains('ref', 9)
    assert mask_subtract.contains('ref', 10)
    assert mask_subtract.contains('ref', 11)
    assert not mask_subtract.contains('ref', 12)
    assert not mask_subtract.contains('ref', 13)
    assert not mask_subtract.contains('ref', 15)

    mask_subtract = mask2.subtract(mask1)
    assert not mask_subtract.contains('ref', 14)
    assert mask_subtract.contains('ref', 15)
    assert mask_subtract.contains('ref', 16)
    assert mask_subtract.contains('ref', 19)
    assert not mask_subtract.contains('ref', 20)

    mask_subtract = mask1.subtract(mask3)
    assert not mask_subtract.contains('ref', 9)
    assert not mask_subtract.contains('ref', 10)
    assert mask_subtract.contains('ref', 11)
    assert mask_subtract.contains('ref', 13)
    assert not mask_subtract.contains('ref', 14)
    assert not mask_subtract.contains('ref', 15)

    mask_subtract = mask1.subtract(mask4)
    assert not mask_subtract.contains('ref', 9)
    assert not mask_subtract.contains('ref', 10)
    assert mask_subtract.contains('ref', 11)
    assert mask_subtract.contains('ref', 13)
    assert mask_subtract.contains('ref', 14)
    assert not mask_subtract.contains('ref', 15)

    mask_subtract = mask1.subtract(mask_empty)
    assert not mask_subtract.contains('ref', 9)
    assert mask_subtract.contains('ref', 10)
    assert mask_subtract.contains('ref', 11)
    assert mask_subtract.contains('ref', 13)
    assert mask_subtract.contains('ref', 14)
    assert not mask_subtract.contains('ref', 15)

    mask_subtract = mask_empty.subtract(mask1)
    assert mask_subtract.is_empty()

    mask_subtract = mask_empty.subtract(mask_empty)
    assert mask_subtract.is_empty()


def test_write():
    sequences = [SeqRecord(seq=Seq('ATCG-NN'), id='record1')]
    mask = MaskedGenomicRegions.from_sequences(sequences=sequences)

    assert 3 == len(mask), 'Invalid length'
    assert not mask.contains('record1', 3)
    assert mask.contains('record1', 4)
    assert mask.contains('record1', 6)
    assert not mask.contains('record1', 7)

    with tempfile.NamedTemporaryFile() as f:
        file = Path(f.name)
        mask.write(file)

        mask2 = MaskedGenomicRegions.from_file(file)
        assert 3 == len(mask2), 'Invalid length'
        assert not mask2.contains('record1', 3)
        assert mask2.contains('record1', 4)
        assert mask2.contains('record1', 6)
        assert not mask2.contains('record1', 7)


def test_mask_genome():
    with tempfile.TemporaryDirectory() as tmp_dir:
        genome_file = Path(tmp_dir) / 'genome.fasta'
        genome_record = SeqRecord(id='reference', seq=Seq('ATCGAATT'))
        with open(genome_file, 'w') as f:
            SeqIO.write(genome_record, f, 'fasta')

        mask = MaskedGenomicRegions(BedTool('reference 0 4', from_string=True))
        masked_records = mask.mask_genome(genome_file, mask_char='?', remove=True)

        assert 1 == len(masked_records)
        assert 'reference' == masked_records['reference'].id
        assert Seq('AATT') == masked_records['reference'].seq


def test_mask_genome_no_remove():
    with tempfile.TemporaryDirectory() as tmp_dir:
        genome_file = Path(tmp_dir) / 'genome.fasta'
        genome_record = SeqRecord(id='reference', seq=Seq('ATCGAATT'))
        with open(genome_file, 'w') as f:
            SeqIO.write(genome_record, f, 'fasta')

        mask = MaskedGenomicRegions(BedTool('reference 0 4', from_string=True))
        masked_records = mask.mask_genome(genome_file, mask_char='?', remove=False)

        assert 1 == len(masked_records)
        assert 'reference' == masked_records['reference'].id
        assert Seq('????AATT') == masked_records['reference'].seq


def test_mask_genome_multiple_sequence():
    with tempfile.TemporaryDirectory() as tmp_dir:
        genome_file = Path(tmp_dir) / 'genome.fasta'
        genome_records = [
            SeqRecord(id='r1', seq=Seq('ATCGAATT')),
            SeqRecord(id='r2', seq=Seq('GGGGCCCCAAAA')),
        ]
        with open(genome_file, 'w') as f:
            SeqIO.write(genome_records, f, 'fasta')

        mask = MaskedGenomicRegions(BedTool('r1 2 4\nr2 4 8', from_string=True))
        masked_records = mask.mask_genome(genome_file, mask_char='?', remove=True)

        assert 2 == len(masked_records)
        assert 'r1' == masked_records['r1'].id
        assert Seq('ATAATT') == masked_records['r1'].seq
        assert 'r2' == masked_records['r2'].id
        assert Seq('GGGGAAAA') == masked_records['r2'].seq


def test_overlaps_region_one_region():
    masked_region = MaskedGenomicRegions(BedTool([('ref', 10, 20)]))

    assert not masked_region.overlaps_range('ref', 9, 10)

    assert masked_region.overlaps_range('ref', 10, 11)
    assert masked_region.overlaps_range('ref', 11, 15)
    assert masked_region.overlaps_range('ref', 11, 21)
    assert masked_region.overlaps_range('ref', 11, 25)

    assert masked_region.overlaps_range('ref', 19, 25)
    assert not masked_region.overlaps_range('ref', 20, 25)
    assert not masked_region.overlaps_range('ref', 21, 25)

    assert masked_region.overlaps_range('ref', 9, 25)

    assert not masked_region.overlaps_range('no_exist', 11, 15)


def test_overlaps_region_multiple_regions():
    masked_region = MaskedGenomicRegions(BedTool([('ref', 10, 20),
                                                  ('ref2', 30, 40)]))

    assert not masked_region.overlaps_range('ref', 9, 10)
    assert not masked_region.overlaps_range('ref', 10, 11, start_position_index='1')
    assert masked_region.overlaps_range('ref', 10, 11)
    assert masked_region.overlaps_range('ref', 11, 12, start_position_index='1')
    assert masked_region.overlaps_range('ref', 11, 15)
    assert masked_region.overlaps_range('ref', 19, 25)
    assert not masked_region.overlaps_range('ref', 20, 25)

    assert not masked_region.overlaps_range('ref2', 29, 30)
    assert masked_region.overlaps_range('ref2', 30, 31)
    assert masked_region.overlaps_range('ref2', 31, 35)
    assert masked_region.overlaps_range('ref2', 39, 45)
    assert not masked_region.overlaps_range('ref2', 40, 45)


def test_positions_iter():
    masked_region = MaskedGenomicRegions(BedTool([('ref', 0, 5)]))

    # 0-based coordinates (default)
    masked_positions = list(masked_region.positions_iter())
    assert masked_positions == [('ref', 0), ('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4)]

    # 1-based coordinates
    masked_positions = list(masked_region.positions_iter(start_position_index='1'))
    assert masked_positions == [('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4), ('ref', 5)]


def test_positions_iter_empty():
    masked_region = MaskedGenomicRegions.empty_mask()

    # 0-based coordinates (default)
    masked_positions = list(masked_region.positions_iter())
    assert masked_positions == []

    # 1-based coordinates
    masked_positions = list(masked_region.positions_iter(start_position_index='1'))
    assert masked_positions == []


def test_positions_iter_multiple_sequences():
    masked_region = MaskedGenomicRegions(BedTool([('ref', 0, 5), ('ref2', 1, 3)]))

    # 0-based coordinates (default)
    masked_positions = list(masked_region.positions_iter())
    assert len(masked_positions) == 7
    assert set(masked_positions) == {('ref', 0), ('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4),
                                     ('ref2', 1), ('ref2', 2)}

    # 1-based coordinates
    masked_positions = list(masked_region.positions_iter(start_position_index='1'))
    assert len(masked_positions) == 7
    assert set(masked_positions) == {('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4), ('ref', 5),
                                     ('ref2', 2), ('ref2', 3)}


def test_positions_iter_multiple_sequences_one_empty():
    masked_region = MaskedGenomicRegions(BedTool([('ref', 0, 5), ('ref2', 1, 1)]))

    # 0-based coordinates (default)
    masked_positions = list(masked_region.positions_iter())
    assert len(masked_positions) == 5
    assert set(masked_positions) == {('ref', 0), ('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4)}

    # 1-based coordinates
    masked_positions = list(masked_region.positions_iter(start_position_index='1'))
    assert len(masked_positions) == 5
    assert set(masked_positions) == {('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4), ('ref', 5)}


def test_positions_iter_multiple_ranges():
    masked_region = MaskedGenomicRegions(BedTool([('ref', 0, 5), ('ref', 10, 13)]))

    # 0-based coordinates (default)
    masked_positions = list(masked_region.positions_iter())
    assert masked_positions == [('ref', 0), ('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4),
                                ('ref', 10), ('ref', 11), ('ref', 12)]

    # 1-based coordinates
    masked_positions = list(masked_region.positions_iter(start_position_index='1'))
    assert masked_positions == [('ref', 1), ('ref', 2), ('ref', 3), ('ref', 4), ('ref', 5),
                                ('ref', 11), ('ref', 12), ('ref', 13)]



import pytest

from genomics_data_index.storage.model.QueryFeatureFactory import QueryFeatureFactory
from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS
from genomics_data_index.storage.model.QueryFeatureHGVSGN import QueryFeatureHGVSGN
from genomics_data_index.storage.model.QueryFeatureMLST import QueryFeatureMLST
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI

query_feature_factory = QueryFeatureFactory.instance()


def test_create_hgvs():
    f = query_feature_factory.create_feature('hgvs:NC_011083:SEHA_RS01180:c.497C>A')
    assert isinstance(f, QueryFeatureHGVS)
    assert 'hgvs:NC_011083:SEHA_RS01180:c.497C>A' == f.id
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'SEHA_RS01180' == f.gene
    assert 'c.497C>A' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert f.is_nucleotide()
    assert not f.is_protein()

    f = query_feature_factory.create_feature('hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu')
    assert isinstance(f, QueryFeatureHGVS)
    assert 'hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu' == f.id
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'SEHA_RS01180' == f.gene
    assert 'p.Ala166Glu' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert not f.is_nucleotide()
    assert f.is_protein()


def test_create_hgvs_gn():
    f = query_feature_factory.create_feature('hgvs_gn:NC_011083:murF:c.497C>A')
    assert isinstance(f, QueryFeatureHGVSGN)
    assert 'hgvs_gn:NC_011083:murF:c.497C>A' == f.id
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'murF' == f.gene
    assert 'c.497C>A' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert f.is_nucleotide()
    assert not f.is_protein()

    f = query_feature_factory.create_feature('hgvs_gn:NC_011083:murF:p.Ala166Glu')
    assert isinstance(f, QueryFeatureHGVSGN)
    assert 'hgvs_gn:NC_011083:murF:p.Ala166Glu' == f.id
    assert 'NC_011083' == f.sequence
    assert 'NC_011083' == f.scope
    assert 'murF' == f.gene
    assert 'p.Ala166Glu' == f.variant
    assert f.has_id()
    assert f.has_gene()
    assert not f.is_nucleotide()
    assert f.is_protein()


def test_create_spdi():
    # Test SNV
    f = query_feature_factory.create_feature('reference:10:A:C')
    assert isinstance(f, QueryFeatureMutationSPDI)
    assert 'reference:10:A:C' == f.id
    assert 'reference' == f.sequence
    assert 10 == f.position
    assert 'A' == f.deletion
    assert 'C' == f.insertion
    assert 1 == f.deletion_length()
    assert f.has_deletion_sequence()
    assert not f.is_unknown()

    # Test complex
    f = query_feature_factory.create_feature('reference:10:ATT:CCC')
    assert isinstance(f, QueryFeatureMutationSPDI)
    assert 'reference:10:ATT:CCC' == f.id
    assert 'reference' == f.sequence
    assert 10 == f.position
    assert 'ATT' == f.deletion
    assert 'CCC' == f.insertion
    assert 3 == f.deletion_length()
    assert f.has_deletion_sequence()
    assert not f.is_unknown()

    # Test complex with integer deletion
    f = query_feature_factory.create_feature('reference:10:3:CCC')
    assert isinstance(f, QueryFeatureMutationSPDI)
    assert 'reference:10:3:CCC' == f.id
    assert 'reference' == f.sequence
    assert 10 == f.position
    assert 3 == f.deletion
    assert 'CCC' == f.insertion
    assert 3 == f.deletion_length()
    assert not f.has_deletion_sequence()
    assert not f.is_unknown()

    # Test unknown integer deletion
    f = query_feature_factory.create_feature('reference:10:1:?')
    assert isinstance(f, QueryFeatureMutationSPDI)
    assert 'reference:10:1:?' == f.id
    assert 'reference' == f.sequence
    assert 10 == f.position
    assert 1 == f.deletion
    assert '?' == f.insertion
    assert 1 == f.deletion_length()
    assert not f.has_deletion_sequence()
    assert f.is_unknown()

    # Test unknown, deletion sequence
    f = query_feature_factory.create_feature('reference:10:A:?')
    assert isinstance(f, QueryFeatureMutationSPDI)
    assert 'reference:10:A:?' == f.id
    assert 'reference' == f.sequence
    assert 10 == f.position
    assert 'A' == f.deletion
    assert '?' == f.insertion
    assert 1 == f.deletion_length()
    assert f.has_deletion_sequence()
    assert f.is_unknown()


def test_create_mlst():
    f = query_feature_factory.create_feature('mlst:ecoli:adk:100')
    assert isinstance(f, QueryFeatureMLST)
    assert 'mlst:ecoli:adk:100' == f.id
    assert 'ecoli:adk:100' == f.id_no_prefix
    assert 'ecoli' == f.scheme
    assert 'ecoli' == f.scope
    assert 'adk' == f.locus
    assert '100' == f.allele
    assert not f.is_unknown()

    f = query_feature_factory.create_feature('mlst:ecoli:adk:?')
    assert isinstance(f, QueryFeatureMLST)
    assert 'mlst:ecoli:adk:?' == f.id
    assert 'ecoli:adk:?' == f.id_no_prefix
    assert 'ecoli' == f.scheme
    assert 'ecoli' == f.scope
    assert 'adk' == f.locus
    assert '?' == f.allele
    assert f.is_unknown()


def test_create_unknown():
    with pytest.raises(Exception) as execinfo:
        query_feature_factory.create_feature('unknown:feature')

    assert 'Unknown feature type for feature=[unknown:feature]' in str(execinfo)



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict

import pytest
from pybedtools import BedTool

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.NucleotideSampleDataSequenceMask import NucleotideSampleDataSequenceMask
from genomics_data_index.test.integration import regular_vcf_dir, data_dir
from genomics_data_index.test.integration.storage.io import read_vcf_df
from genomics_data_index.test.integration.storage.io.mutation import vcf_and_bed_mask_files


@pytest.fixture
def files_map():
    return {
        'SampleA': {
            'vcf': Path(regular_vcf_dir, 'SampleA.vcf.gz'),
            'mask': Path(data_dir, 'SampleA', 'snps.aligned.fa'),
        },
        'SampleB': {
            'vcf': Path(regular_vcf_dir, 'SampleB.vcf.gz'),
            'mask': Path(data_dir, 'SampleB', 'snps.aligned.fa'),
        },
        'SampleC': {
            'vcf': Path(regular_vcf_dir, 'SampleC.vcf.gz'),
            'mask': Path(data_dir, 'SampleC', 'snps.aligned.fa'),
        }
    }


@pytest.fixture
def files_bed_mask_map(sample_dirs):
    return vcf_and_bed_mask_files(sample_dirs)


def test_get_vcf_file_no_preprocess(files_map: Dict[str, Dict[str, Path]]):
    sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                           vcf_file=files_map['SampleA']['vcf'],
                                                           sample_mask_sequence=files_map['SampleA']['mask'])
    vcf_file, index_file = sample_files.get_vcf_file(ignore_preprocessed=True)
    assert vcf_file.exists()
    assert index_file is None
    df = read_vcf_df(vcf_file)
    assert len(df[df['TYPE'].isna()]) > 0


def test_get_sample_mask_no_preprocess(files_map: Dict[str, Dict[str, Path]]):
    sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                           vcf_file=files_map['SampleA']['vcf'],
                                                           sample_mask_sequence=files_map['SampleA']['mask'])
    with pytest.raises(Exception) as execinfo:
        sample_files.get_mask()
    assert 'Sample mask file is not preprocessed for sample' in str(execinfo.value)


def test_get_empty_sample_mask_no_preprocess(files_map: Dict[str, Dict[str, Path]]):
    sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                           vcf_file=files_map['SampleA']['vcf'],
                                                           sample_mask_sequence=None)
    with pytest.raises(Exception) as execinfo:
        sample_files.get_mask()
    assert 'Sample mask file is not preprocessed for sample' in str(execinfo.value)


def test_get_sample_mask_after_preprocess(files_map: Dict[str, Dict[str, Path]]):
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                               vcf_file=files_map['SampleA']['vcf'],
                                                               sample_mask_sequence=files_map['SampleA']['mask'])
        assert not sample_files.is_preprocessed()
        processed_sample_files = sample_files.persist(tmp_dir)
        assert processed_sample_files.is_preprocessed()
        mask = processed_sample_files.get_mask()
        assert 437 == len(mask)
        assert {'reference'} == mask.sequence_names()
        mask_file = processed_sample_files.get_mask_file()
        assert mask_file.exists()
        assert mask_file.parent == tmp_dir


def test_get_sample_mask_after_preprocess_subtract_vcf(files_map: Dict[str, Dict[str, Path]],
                                                       files_bed_mask_map: Dict[str, Dict[str, Path]]):
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                               vcf_file=files_map['SampleA']['vcf'],
                                                               sample_mask_sequence=files_map['SampleA']['mask'],
                                                               subtract_vcf_from_mask=True)
        expected_bed = BedTool(str(files_bed_mask_map['masks_minus_vcf']['SampleA']))
        expected_mask = MaskedGenomicRegions(expected_bed)

        assert not sample_files.is_preprocessed()
        processed_sample_files = sample_files.persist(tmp_dir)
        assert processed_sample_files.is_preprocessed()
        mask = processed_sample_files.get_mask()
        assert expected_mask.mask == mask.mask, f'expected=\n{expected_mask.mask} != actual=\n{mask.mask}'
        assert 414 == len(mask)
        assert {'reference'} == mask.sequence_names()
        mask_file = processed_sample_files.get_mask_file()
        assert mask_file.exists()
        assert mask_file.parent == tmp_dir


def test_get_sample_empty_mask_after_preprocess(files_map: Dict[str, Dict[str, Path]]):
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                               vcf_file=files_map['SampleA']['vcf'])
        assert not sample_files.is_preprocessed()
        processed_sample_files = sample_files.persist(tmp_dir)
        assert processed_sample_files.is_preprocessed()
        mask_file = processed_sample_files.get_mask_file()
        assert mask_file.exists()
        assert mask_file.parent == tmp_dir
        mask = processed_sample_files.get_mask()
        assert mask.is_empty()


def test_get_sample_mask_double_preprocess(files_map: Dict[str, Dict[str, Path]]):
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                               vcf_file=files_map['SampleA']['vcf'],
                                                               sample_mask_sequence=files_map['SampleA']['mask'])
        assert not sample_files.is_preprocessed()
        processed_sample_files = sample_files.persist(tmp_dir)
        assert processed_sample_files.is_preprocessed()
        mask = processed_sample_files.get_mask()
        assert 437 == len(mask)
        assert {'reference'} == mask.sequence_names()
        mask_file = processed_sample_files.get_mask_file()

        with TemporaryDirectory() as tmp_dir_2_str:
            tmp_dir_2 = Path(tmp_dir_2_str)
            processed_sample_files_2 = processed_sample_files.persist(tmp_dir_2)
            assert processed_sample_files_2.is_preprocessed()
            mask_2 = processed_sample_files_2.get_mask()
            assert 437 == len(mask)
            assert {'reference'} == mask_2.sequence_names()
            mask_file_2 = processed_sample_files_2.get_mask_file()

            assert mask_file.exists()
            assert mask_file_2.exists()
            assert mask_file.parent == tmp_dir
            assert mask_file_2.parent == tmp_dir_2


def test_get_vcf_file_no_preprocess_exception(files_map: Dict[str, Dict[str, Path]]):
    sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                           vcf_file=files_map['SampleA']['vcf'],
                                                           sample_mask_sequence=files_map['SampleA']['mask'])
    with pytest.raises(Exception) as execinfo:
        sample_files.get_vcf_file()
    assert 'VCF file for sample' in str(execinfo.value)


def test_get_vcf_file_with_preprocess(files_map: Dict[str, Dict[str, Path]]):
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                               vcf_file=files_map['SampleA']['vcf'],
                                                               sample_mask_sequence=files_map['SampleA']['mask'])
        assert not sample_files.is_preprocessed()
        with pytest.raises(Exception) as execinfo:
            sample_files.get_vcf_file()
        assert 'VCF file for sample' in str(execinfo.value)

        processed_sample_files = sample_files.persist(tmp_dir)
        assert processed_sample_files.is_preprocessed()

        vcf_file, index_file = processed_sample_files.get_vcf_file()
        assert vcf_file.exists()
        assert index_file.exists()
        assert vcf_file.parent == tmp_dir
        assert index_file.parent == tmp_dir

        df = read_vcf_df(vcf_file)
        assert len(df[df['TYPE'].isna()]) == 0
        assert 'SNP' == df[df['POS'] == 293]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 302]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 324]['TYPE'].tolist()[0]
        assert 'INDEL' == df[df['POS'] == 374]['TYPE'].tolist()[0]
        assert 'OTHER' == df[df['POS'] == 461]['TYPE'].tolist()[0]
        assert 'SNP' == df[df['POS'] == 506]['TYPE'].tolist()[0]


def test_get_vcf_file_double_preprocess(files_map: Dict[str, Dict[str, Path]]):
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_files = NucleotideSampleDataSequenceMask.create(sample_name='SampleA',
                                                               vcf_file=files_map['SampleA']['vcf'],
                                                               sample_mask_sequence=files_map['SampleA']['mask'])
        assert not sample_files.is_preprocessed()
        processed_sample_files = sample_files.persist(tmp_dir)
        assert processed_sample_files.is_preprocessed()
        vcf_file, index_file = processed_sample_files.get_vcf_file()
        assert vcf_file.exists()
        assert index_file.exists()

        with TemporaryDirectory() as tmp_dir_2_str:
            tmp_dir_2 = Path(tmp_dir_2_str)
            processed_sample_files_2 = processed_sample_files.persist(tmp_dir_2)
            assert processed_sample_files_2.is_preprocessed()
            vcf_file_2, index_file_2 = processed_sample_files_2.get_vcf_file()
            assert vcf_file_2.exists()
            assert index_file_2.exists()

            assert vcf_file.parent == tmp_dir
            assert index_file.parent == tmp_dir
            assert vcf_file_2.parent == tmp_dir_2
            assert index_file_2.parent == tmp_dir_2



from __future__ import annotations

import abc
from typing import Generator, Set

from genomics_data_index.storage.io.FeaturesReader import FeaturesReader
from genomics_data_index.storage.io.SampleData import SampleData


class SampleDataPackage(abc.ABC):

    def __init__(self, index_unknown_missing: bool):
        self._index_unknown_missing = index_unknown_missing

    @abc.abstractmethod
    def sample_names(self) -> Set[str]:
        pass

    def index_unknown_missing(self) -> bool:
        return self._index_unknown_missing

    @abc.abstractmethod
    def process_all_data(self) -> SampleDataPackage:
        pass

    @abc.abstractmethod
    def get_features_reader(self) -> FeaturesReader:
        pass

    @abc.abstractmethod
    def iter_sample_data(self) -> Generator[SampleData, None, None]:
        pass



#!/usr/bin/env python
# Used to print out unknown/missing ('N' or '-') in snippy alignment files for debugging

import sys

from Bio import SeqIO

if len(sys.argv) != 2:
    print(f"Usage: {sys.argv[0]} [snps.aligned.fa]\n")
    sys.exit(1)

with open(sys.argv[1]) as handle:
    for record in SeqIO.parse(handle, "fasta"):
        name = record.id
        for i, c in enumerate(record.seq):
            c = c.upper()
            if c == "N" or c == "-":
                pos = i + 1
                print(pos)



from pathlib import Path

from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile

# Prepares a SnpEff database from a reference genome

input_reference = Path(snakemake.input.reference)
output_db_conf = Path(snakemake.output.snpeff_db_conf)
snpeff_db_dir = output_db_conf.parent.absolute()

sequence_file = SequenceFile(input_reference)
snpeff_database = sequence_file.create_snpeff_database(database_dir=snpeff_db_dir)
snpeff_database.write_database_config(output_db_conf)



import pandas as pd

from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData


def test_combine_vcf_mask():
    num_annotations = 9

    data_vcf = [
        ['SampleA', 'ref', 10, 'A', 'T', 'SNP', 'file', 'ref:10:A:T'] + [pd.NA] * num_annotations,
    ]
    data_mask = [
        ['SampleA', 'ref', 1, 1, '?', 'UNKNOWN_MISSING', 'file', 'ref:1:1:?']
    ]
    columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
               'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
               'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sample_data = NucleotideSampleData('SampleA', vcf_file=None, vcf_file_index=None,
                                       mask_bed_file=None, preprocessed=True)

    vcf_df = pd.DataFrame(data_vcf, columns=columns)
    mask_df = pd.DataFrame(data_mask, columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID'])

    combined_df = sample_data.combine_vcf_mask(vcf_frame=vcf_df, mask_frame=mask_df)
    combined_df = combined_df.sort_values('POS').fillna('NA')

    assert combined_df.columns.tolist() == columns
    assert 2 == len(combined_df)
    assert ['ref', 'ref'] == combined_df['CHROM'].tolist()
    assert [1, 10] == combined_df['POS'].tolist()
    assert [1, 'A'] == combined_df['REF'].tolist()
    assert ['?', 'T'] == combined_df['ALT'].tolist()
    assert ['UNKNOWN_MISSING', 'SNP'] == combined_df['TYPE'].tolist()
    assert ['ref:1:1:?', 'ref:10:A:T'] == combined_df['VARIANT_ID'].tolist()
    assert ['NA', 'NA'] == combined_df['ANN.Annotation'].tolist()


def test_combine_vcf_mask_overlap_feature():
    num_annotations = 9

    data_vcf = [
        ['SampleA', 'ref', 10, 'A', 'T', 'SNP', 'file', 'ref:10:A:T'] + [pd.NA] * num_annotations,
    ]
    data_mask = [
        ['SampleA', 'ref', 10, 1, '?', 'UNKNOWN_MISSING', 'file', 'ref:10:1:?']
    ]
    columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
               'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
               'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sample_data = NucleotideSampleData('SampleA', vcf_file=None, vcf_file_index=None,
                                       mask_bed_file=None, preprocessed=True)

    vcf_df = pd.DataFrame(data_vcf, columns=columns)
    mask_df = pd.DataFrame(data_mask, columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID'])

    combined_df = sample_data.combine_vcf_mask(vcf_frame=vcf_df, mask_frame=mask_df)
    combined_df = combined_df.sort_values('POS').fillna('NA')

    assert combined_df.columns.tolist() == columns
    assert 1 == len(combined_df)
    assert ['ref'] == combined_df['CHROM'].tolist()
    assert [10] == combined_df['POS'].tolist()
    assert [1] == combined_df['REF'].tolist()
    assert ['?'] == combined_df['ALT'].tolist()
    assert ['UNKNOWN_MISSING'] == combined_df['TYPE'].tolist()
    assert ['ref:10:1:?'] == combined_df['VARIANT_ID'].tolist()
    assert ['NA'] == combined_df['ANN.Annotation'].tolist()


def test_combine_vcf_mask_no_mask_features():
    num_annotations = 9

    data_vcf = [
        ['SampleA', 'ref', 10, 'A', 'T', 'SNP', 'file', 'ref:10:A:T'] + [pd.NA] * num_annotations,
    ]
    data_mask = [
    ]
    columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
               'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
               'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sample_data = NucleotideSampleData('SampleA', vcf_file=None, vcf_file_index=None,
                                       mask_bed_file=None, preprocessed=True)

    vcf_df = pd.DataFrame(data_vcf, columns=columns)
    mask_df = pd.DataFrame(data_mask, columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID'])

    combined_df = sample_data.combine_vcf_mask(vcf_frame=vcf_df, mask_frame=mask_df)
    combined_df = combined_df.sort_values('POS').fillna('NA')

    assert combined_df.columns.tolist() == columns
    assert 1 == len(combined_df)
    assert ['ref'] == combined_df['CHROM'].tolist()
    assert [10] == combined_df['POS'].tolist()
    assert ['A'] == combined_df['REF'].tolist()
    assert ['T'] == combined_df['ALT'].tolist()
    assert ['SNP'] == combined_df['TYPE'].tolist()
    assert ['ref:10:A:T'] == combined_df['VARIANT_ID'].tolist()
    assert ['NA'] == combined_df['ANN.Annotation'].tolist()


def test_combine_vcf_mask_no_vcf_feature():
    data_vcf = [
    ]
    data_mask = [
        ['SampleA', 'ref', 10, 1, '?', 'UNKNOWN_MISSING', 'file', 'ref:10:1:?']
    ]
    columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
               'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
               'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sample_data = NucleotideSampleData('SampleA', vcf_file=None, vcf_file_index=None,
                                       mask_bed_file=None, preprocessed=True)

    vcf_df = pd.DataFrame(data_vcf, columns=columns)
    mask_df = pd.DataFrame(data_mask, columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID'])

    combined_df = sample_data.combine_vcf_mask(vcf_frame=vcf_df, mask_frame=mask_df)
    combined_df = combined_df.sort_values('POS').fillna('NA')

    assert combined_df.columns.tolist() == columns
    assert 1 == len(combined_df)
    assert ['ref'] == combined_df['CHROM'].tolist()
    assert [10] == combined_df['POS'].tolist()
    assert [1] == combined_df['REF'].tolist()
    assert ['?'] == combined_df['ALT'].tolist()
    assert ['UNKNOWN_MISSING'] == combined_df['TYPE'].tolist()
    assert ['ref:10:1:?'] == combined_df['VARIANT_ID'].tolist()
    assert ['NA'] == combined_df['ANN.Annotation'].tolist()


def test_combine_vcf_mask_same_position_different_sequence():
    num_annotations = 9

    data_vcf = [
        ['SampleA', 'ref', 10, 'A', 'T', 'SNP', 'file', 'ref:10:A:T'] + [pd.NA] * num_annotations,
    ]
    data_mask = [
        ['SampleA', 'ref2', 10, 1, '?', 'UNKNOWN_MISSING', 'file', 'ref2:10:1:?']
    ]
    columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
               'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
               'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sample_data = NucleotideSampleData('SampleA', vcf_file=None, vcf_file_index=None,
                                       mask_bed_file=None, preprocessed=True)

    vcf_df = pd.DataFrame(data_vcf, columns=columns)
    mask_df = pd.DataFrame(data_mask, columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID'])

    combined_df = sample_data.combine_vcf_mask(vcf_frame=vcf_df, mask_frame=mask_df)
    combined_df = combined_df.sort_values(['CHROM', 'POS']).fillna('NA')

    assert combined_df.columns.tolist() == columns
    assert 2 == len(combined_df)
    assert ['ref', 'ref2'] == combined_df['CHROM'].tolist()
    assert [10, 10] == combined_df['POS'].tolist()
    assert ['A', 1] == combined_df['REF'].tolist()
    assert ['T', '?'] == combined_df['ALT'].tolist()
    assert ['SNP', 'UNKNOWN_MISSING'] == combined_df['TYPE'].tolist()
    assert ['ref:10:A:T', 'ref2:10:1:?'] == combined_df['VARIANT_ID'].tolist()
    assert ['NA', 'NA'] == combined_df['ANN.Annotation'].tolist()


def test_combine_vcf_mask_multiple_features():
    num_annotations = 9

    data_vcf = [
        ['SampleA', 'ref', 10, 'A', 'T', 'SNP', 'file', 'ref:10:A:T'] + [pd.NA] * num_annotations,
        ['SampleA', 'ref', 20, 'ATT', 'T', 'INDEL', 'file', 'ref:20:ATT:T'] + [pd.NA] * num_annotations,
    ]
    data_mask = [
        ['SampleA', 'ref', 10, 1, '?', 'UNKNOWN_MISSING', 'file', 'ref:10:1:?'],
        ['SampleA', 'ref', 30, 1, '?', 'UNKNOWN_MISSING', 'file', 'ref:30:1:?']
    ]
    columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
               'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
               'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sample_data = NucleotideSampleData('SampleA', vcf_file=None, vcf_file_index=None,
                                       mask_bed_file=None, preprocessed=True)

    vcf_df = pd.DataFrame(data_vcf, columns=columns)
    mask_df = pd.DataFrame(data_mask, columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID'])

    combined_df = sample_data.combine_vcf_mask(vcf_frame=vcf_df, mask_frame=mask_df)
    combined_df = combined_df.sort_values('POS').fillna('NA')

    assert combined_df.columns.tolist() == columns
    assert 3 == len(combined_df)
    assert ['ref', 'ref', 'ref'] == combined_df['CHROM'].tolist()
    assert [10, 20, 30] == combined_df['POS'].tolist()
    assert [1, 'ATT', 1] == combined_df['REF'].tolist()
    assert ['?', 'T', '?'] == combined_df['ALT'].tolist()
    assert ['UNKNOWN_MISSING', 'INDEL', 'UNKNOWN_MISSING'] == combined_df['TYPE'].tolist()
    assert ['ref:10:1:?', 'ref:20:ATT:T', 'ref:30:1:?'] == combined_df['VARIANT_ID'].tolist()
    assert ['NA', 'NA', 'NA'] == combined_df['ANN.Annotation'].tolist()



import pytest

from genomics_data_index.storage.model.NucleotideMutationTranslater import NucleotideMutationTranslater
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI


def test_to_spdi_position():
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 1, 'T')
    assert 'seq:2:1:T' == NucleotideMutationTranslater.to_spdi('seq', 2, 1, 'T')
    assert 'seq:0:1:T' == NucleotideMutationTranslater.to_spdi('seq', 0, 1, 'T')

    with pytest.raises(Exception) as execinfo:
        NucleotideMutationTranslater.to_spdi('seq', -1, 1, 'T')
    assert 'Position must be non-negative' in str(execinfo.value)


def test_to_spdi_deletion():
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 1, 'T')
    assert 'seq:1:2:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 2, 'T')
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 'A', 'T')
    assert 'seq:1:2:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 'AT', 'T')
    assert 'seq:1:0:T' == NucleotideMutationTranslater.to_spdi('seq', 1, '', 'T')
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, '1', 'T')


def test_to_spdi_deletion_no_convert_deletion():
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 1, 'T', convert_deletion=False)
    assert 'seq:1:2:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 2, 'T', convert_deletion=False)
    assert 'seq:1:A:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 'A', 'T', convert_deletion=False)
    assert 'seq:1:AT:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 'AT', 'T', convert_deletion=False)
    assert 'seq:1::T' == NucleotideMutationTranslater.to_spdi('seq', 1, '', 'T', convert_deletion=False)
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, '1', 'T', convert_deletion=False)


def test_to_spdi_alt():
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_spdi('seq', 1, 1, 'T')
    assert 'seq:1:1:AT' == NucleotideMutationTranslater.to_spdi('seq', 1, 1, 'AT')
    assert 'seq:1:1:' == NucleotideMutationTranslater.to_spdi('seq', 1, 1, '')


def test_from_spdi_position():
    assert ('seq', 1, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:A:T')
    assert ('seq', 2, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:2:A:T')
    assert ('seq', 0, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:0:A:T')

    with pytest.raises(Exception) as execinfo:
        NucleotideMutationTranslater.from_spdi('seq:-1:A:T')
    assert 'Position must be non-negative' in str(execinfo.value)


def test_from_spdi_deletion():
    assert ('seq', 1, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:A:T')
    assert ('seq', 1, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:1:T')
    assert ('seq', 1, 2, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:2:T')
    assert ('seq', 1, 2, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:AT:T')
    assert ('seq', 1, 0, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:0:T')
    assert ('seq', 1, 0, 'T') == NucleotideMutationTranslater.from_spdi('seq:1::T')


def test_from_spdi_deletion_no_convert_deletion():
    assert ('seq', 1, 'A', 'T') == NucleotideMutationTranslater.from_spdi('seq:1:A:T', convert_deletion=False)
    assert ('seq', 1, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:1:T', convert_deletion=False)
    assert ('seq', 1, 2, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:2:T', convert_deletion=False)
    assert ('seq', 1, 'AT', 'T') == NucleotideMutationTranslater.from_spdi('seq:1:AT:T', convert_deletion=False)
    assert ('seq', 1, 0, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:0:T', convert_deletion=False)

    with pytest.raises(Exception) as execinfo:
        NucleotideMutationTranslater.from_spdi('seq:1::T', convert_deletion=False)
    assert 'deletion=[] but convert_deletion is False' in str(execinfo.value)


def test_from_spdi_insertion():
    assert ('seq', 1, 1, 'T') == NucleotideMutationTranslater.from_spdi('seq:1:A:T')
    assert ('seq', 2, 1, 'TA') == NucleotideMutationTranslater.from_spdi('seq:2:A:TA')
    assert ('seq', 0, 1, '') == NucleotideMutationTranslater.from_spdi('seq:0:A:')


def test_to_db_feature():
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_db_feature(QueryFeatureMutationSPDI('seq:1:A:T')).id
    assert 'seq:1:1:T' == NucleotideMutationTranslater.to_db_feature(QueryFeatureMutationSPDI('seq:1:1:T')).id
    assert 'seq:1:2:T' == NucleotideMutationTranslater.to_db_feature(QueryFeatureMutationSPDI('seq:1:AT:T')).id
    assert 'seq:1:2:T' == NucleotideMutationTranslater.to_db_feature(QueryFeatureMutationSPDI('seq:1:2:T')).id
    assert 'seq:10:2:TA' == NucleotideMutationTranslater.to_db_feature(QueryFeatureMutationSPDI('seq:10:AT:TA')).id


def test_to_hgvs_id():
    assert 'hgvs:NC_011083:SEHA_RS01180:p.Ala166Glu' == NucleotideMutationTranslater.to_hgvs_id('NC_011083',
                                                                                                'SEHA_RS01180',
                                                                                                'p.Ala166Glu')
    assert 'hgvs:NC_011083:n.1031571T>C' == NucleotideMutationTranslater.to_hgvs_id('NC_011083', None, 'n.1031571T>C')
    assert 'hgvs:NC_011083:n.1031571T>C' == NucleotideMutationTranslater.to_hgvs_id('NC_011083', 'gene_name',
                                                                                    'n.1031571T>C')



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List

import pandas as pd
import pytest

from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.storage.model.db import NucleotideVariantsSamples, SampleNucleotideVariation, Sample
from genomics_data_index.storage.service import EntityExistsError
from genomics_data_index.storage.service.VariationService import VariationService
from genomics_data_index.test.integration import data_dir, snippy_snps_dataframes, snippy_all_dataframes
from genomics_data_index.test.integration import reference_file_snpeff


def test_insert_variants_saved_files(database, snippy_nucleotide_data_package, reference_service_with_data,
                                     sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=50)

    session = database.get_session()

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    samples = session.query(Sample).all()

    assert {'SampleA', 'SampleB', 'SampleC'} == {s.name for s in samples}

    variation_files = {v.nucleotide_variants_file for s in samples for v in s.sample_nucleotide_variation}
    assert 3 == len(variation_files)

    genomic_mask_files = {v.masked_regions_file for s in samples for v in s.sample_nucleotide_variation}
    assert 3 == len(genomic_mask_files)


def test_insert_variants_saved_files_set_no_index_package(database, snippy_nucleotide_data_package_no_index_missing,
                                                          reference_service_with_data,
                                                          sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)

    session = database.get_session()

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package_no_index_missing)

    samples = session.query(Sample).all()

    assert {'SampleA', 'SampleB', 'SampleC'} == {s.name for s in samples}

    variation_files = {v.nucleotide_variants_file for s in samples for v in s.sample_nucleotide_variation}
    assert 3 == len(variation_files)

    genomic_mask_files = {v.masked_regions_file for s in samples for v in s.sample_nucleotide_variation}
    assert 3 == len(genomic_mask_files)


def test_read_index(database, snippy_nucleotide_data_package, reference_service_with_data,
                    sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    session = database.get_session()
    sampleB = session.query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = session.query(Sample).filter(Sample.name == 'SampleC').one()

    # Read 2 that do exist
    feature_samples = variation_service.read_index(['reference:839:1:G', 'reference:866:9:G'])
    assert len(feature_samples) == 2
    assert {'reference:839:1:G', 'reference:866:9:G'} == set(feature_samples.keys())
    assert isinstance(feature_samples['reference:839:1:G'], NucleotideVariantsSamples)
    assert isinstance(feature_samples['reference:866:9:G'], NucleotideVariantsSamples)
    assert {sampleB.id, sampleC.id} == set(feature_samples['reference:839:1:G'].sample_ids)
    assert {sampleC.id} == set(feature_samples['reference:866:9:G'].sample_ids)

    # Read empty list
    feature_samples = variation_service.read_index([])
    assert len(feature_samples) == 0

    # Read list with additional non-existing IDs
    feature_samples = variation_service.read_index(['reference:839:1:G', 'reference:866:9:G', 'not_exist:100:A:T'])
    assert len(feature_samples) == 2
    assert {'reference:839:1:G', 'reference:866:9:G'} == set(feature_samples.keys())


def test_count_on_reference_no_index_unknowns(database, snippy_nucleotide_data_package, reference_service_with_data,
                                              sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    assert 112 == variation_service.count_on_reference(reference_name='genome', include_unknown=False)
    assert 112 == variation_service.count_on_reference(reference_name='genome', include_unknown=True)

    with pytest.raises(EntityExistsError) as execinfo:
        variation_service.count_on_reference('no_exists')
    assert 'No reference genome with name=[no_exists]' in str(execinfo.value)


def test_count_on_reference_with_index_unknowns_low_slice_limit(database, snippy_nucleotide_data_package,
                                                                reference_service_with_data,
                                                                sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=5)

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    assert 112 == variation_service.count_on_reference(reference_name='genome', include_unknown=False)
    assert 112 + 440 == variation_service.count_on_reference(reference_name='genome', include_unknown=True)

    with pytest.raises(EntityExistsError) as execinfo:
        variation_service.count_on_reference('no_exists')
    assert 'No reference genome with name=[no_exists]' in str(execinfo.value)


def test_count_on_reference_with_index_unknowns_high_slice_limit(database, snippy_nucleotide_data_package,
                                                                 reference_service_with_data,
                                                                 sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=1000)

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    assert 112 == variation_service.count_on_reference(reference_name='genome', include_unknown=False)
    assert 112 + 440 == variation_service.count_on_reference(reference_name='genome', include_unknown=True)

    with pytest.raises(EntityExistsError) as execinfo:
        variation_service.count_on_reference('no_exists')
    assert 'No reference genome with name=[no_exists]' in str(execinfo.value)


def test_insert_variants_masked_regions(database, snippy_nucleotide_data_package, reference_service_with_data,
                                        sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)

    session = database.get_session()

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    samples = session.query(Sample).all()

    genomic_masks = {s.name: v.masked_regions for s in samples for v in s.sample_nucleotide_variation}
    assert 414 == len(genomic_masks['SampleA'])
    assert 234 == len(genomic_masks['SampleB'])
    assert 302 == len(genomic_masks['SampleC'])

    assert 112 + 440 == session.query(NucleotideVariantsSamples).count()
    assert 440 == session.query(NucleotideVariantsSamples) \
        .filter(NucleotideVariantsSamples.var_type == 'UNKNOWN_MISSING') \
        .count()
    assert 112 == session.query(NucleotideVariantsSamples) \
        .filter(NucleotideVariantsSamples.var_type != 'UNKNOWN_MISSING') \
        .count()


def test_mutation_counts_on_reference(database, snippy_nucleotide_data_package, reference_service_with_data,
                                      sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    mutation_counts = variation_service.mutation_counts_on_reference('genome')
    assert 112 == len(mutation_counts)
    assert 2 == mutation_counts['reference:839:1:G']
    assert 1 == mutation_counts['reference:866:9:G']
    assert 1 == mutation_counts['reference:1048:1:G']
    assert 2 == mutation_counts['reference:3897:5:G']

    mutation_counts = variation_service.mutation_counts_on_reference('genome', include_unknown=True)
    assert 112 + 440 == len(mutation_counts)
    assert 2 == mutation_counts['reference:839:1:G']
    assert 1 == mutation_counts['reference:866:9:G']
    assert 1 == mutation_counts['reference:1048:1:G']
    assert 2 == mutation_counts['reference:3897:5:G']
    assert 2 == mutation_counts['reference:3897:5:G']

    assert 3 == mutation_counts['reference:9:1:?']
    assert 2 == mutation_counts['reference:5100:1:?']
    assert 'reference:888:1:?' not in mutation_counts
    assert 'reference:302:1:?' not in mutation_counts
    assert 'reference:303:1:?' not in mutation_counts


def test_get_variants_on_reference_no_index_unknowns(database, snippy_nucleotide_data_package,
                                                     reference_service_with_data,
                                                     sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    mutations = variation_service.get_variants_on_reference('genome', include_unknown=False)
    assert 112 == len(mutations)
    assert 2 == len(mutations['reference:839:1:G'].sample_ids)
    assert 839 == mutations['reference:839:1:G'].position

    assert 1 == len(mutations['reference:866:9:G'].sample_ids)
    assert 866 == mutations['reference:866:9:G'].position

    assert 1 == len(mutations['reference:1048:1:G'].sample_ids)
    assert 1048 == mutations['reference:1048:1:G'].position

    assert 2 == len(mutations['reference:3897:5:G'].sample_ids)
    assert 3897 == mutations['reference:3897:5:G'].position

    # Test should still have the same number of mutations when including unknowns since none were indexed
    mutations = variation_service.get_variants_on_reference('genome', include_unknown=True)
    assert 112 == len(mutations)
    assert 'reference:839:1:G' in mutations

    # Test setting include_present to False, which should give no results in this case
    mutations = variation_service.get_variants_on_reference('genome', include_present=False, include_unknown=True)
    assert 0 == len(mutations)


def test_get_variants_on_reference_index_unknowns(database, snippy_nucleotide_data_package, reference_service_with_data,
                                                  sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    mutations = variation_service.get_variants_on_reference('genome', include_unknown=True)
    assert 112 + 440 == len(mutations)
    assert 2 == len(mutations['reference:839:1:G'].sample_ids)
    assert 839 == mutations['reference:839:1:G'].position

    assert 1 == len(mutations['reference:866:9:G'].sample_ids)
    assert 866 == mutations['reference:866:9:G'].position

    assert 1 == len(mutations['reference:1048:1:G'].sample_ids)
    assert 1048 == mutations['reference:1048:1:G'].position

    assert 2 == len(mutations['reference:3897:5:G'].sample_ids)
    assert 3897 == mutations['reference:3897:5:G'].position

    assert 3 == len(mutations['reference:87:1:?'].sample_ids)
    assert 87 == mutations['reference:87:1:?'].position

    mutations = variation_service.get_variants_on_reference('genome', include_unknown=False)
    assert 112 == len(mutations)
    assert 2 == len(mutations['reference:839:1:G'].sample_ids)
    assert 839 == mutations['reference:839:1:G'].position

    assert 1 == len(mutations['reference:866:9:G'].sample_ids)
    assert 866 == mutations['reference:866:9:G'].position

    assert 1 == len(mutations['reference:1048:1:G'].sample_ids)
    assert 1048 == mutations['reference:1048:1:G'].position

    assert 2 == len(mutations['reference:3897:5:G'].sample_ids)
    assert 3897 == mutations['reference:3897:5:G'].position

    assert 'reference:87:1:?' not in mutations

    # Test only including unknowns
    mutations = variation_service.get_variants_on_reference('genome', include_present=False, include_unknown=True)
    assert 440 == len(mutations)
    assert 'reference:87:1:?' in mutations
    assert 'reference:839:1:G' not in mutations

    # Test inluding neigther present nor unknowns
    mutations = variation_service.get_variants_on_reference('genome', include_present=False, include_unknown=False)
    assert 0 == len(mutations)


def test_get_features(database, snippy_nucleotide_data_package, reference_service_with_data,
                      sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    # Test include unknown
    mutations = variation_service.get_features(include_unknown=True)
    assert 112 + 440 == len(mutations)
    assert 2 == len(mutations['reference:839:1:G'].sample_ids)
    assert 1 == len(mutations['reference:866:9:G'].sample_ids)
    assert 1 == len(mutations['reference:1048:1:G'].sample_ids)
    assert 2 == len(mutations['reference:3897:5:G'].sample_ids)
    assert 3 == len(mutations['reference:87:1:?'].sample_ids)

    # Test include unknown ids translated
    mutations = variation_service.get_features(include_unknown=True, id_type='spdi_ref')
    assert 112 + 440 == len(mutations)
    assert 2 == len(mutations['reference:839:C:G'].sample_ids)
    assert 1 == len(mutations['reference:866:GCCAGATCC:G'].sample_ids)
    assert 1 == len(mutations['reference:1048:C:G'].sample_ids)
    assert 2 == len(mutations['reference:3897:GCGCA:G'].sample_ids)
    assert 3 == len(mutations['reference:87:G:?'].sample_ids)

    # Test no include unknown
    mutations = variation_service.get_features(include_unknown=False)
    assert 112 == len(mutations)
    assert 2 == len(mutations['reference:839:1:G'].sample_ids)
    assert 1 == len(mutations['reference:866:9:G'].sample_ids)
    assert 1 == len(mutations['reference:1048:1:G'].sample_ids)
    assert 2 == len(mutations['reference:3897:5:G'].sample_ids)
    assert 'reference:87:1:?' not in mutations

    # Test no include unknown ids translated
    mutations = variation_service.get_features(include_unknown=False,
                                               id_type='spdi_ref')
    assert 112 == len(mutations)
    assert 2 == len(mutations['reference:839:C:G'].sample_ids)
    assert 1 == len(mutations['reference:866:GCCAGATCC:G'].sample_ids)
    assert 1 == len(mutations['reference:1048:C:G'].sample_ids)
    assert 2 == len(mutations['reference:3897:GCGCA:G'].sample_ids)
    assert 'reference:87:G:?' not in mutations
    assert 'reference:87:1:?' not in mutations

    # Test only including unknowns
    mutations = variation_service.get_features(include_present=False, include_unknown=True)
    assert 440 == len(mutations)
    assert 3 == len(mutations['reference:87:1:?'].sample_ids)
    assert 'reference:839:1:G' not in mutations

    # Test only including unknowns, id translated
    mutations = variation_service.get_features(include_present=False, include_unknown=True,
                                               id_type='spdi_ref')
    assert 440 == len(mutations)
    assert 3 == len(mutations['reference:87:G:?'].sample_ids)
    assert 'reference:839:1:G' not in mutations

    # Test including neither present nor unknowns
    mutations = variation_service.get_features(include_present=False, include_unknown=False)
    assert 0 == len(mutations)


def test_insert_variants_examine_variation_with_unknown(database, snippy_nucleotide_data_package,
                                                        reference_service_with_data,
                                                        sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    session = database.get_session()

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    sample_name_ids = sample_service.find_sample_name_ids({'SampleA', 'SampleB', 'SampleC'})
    assert 3 == len(sample_name_ids.values())

    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'

    assert 112 + 440 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'

    # Check to make sure some variants are stored
    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1048,
        'deletion': len('C'),
        'insertion': 'G'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleA']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1135,
        'deletion': len('CCT'),
        'insertion': 'C'
    })
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SampleB'], sample_name_ids['SampleC']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 883,
        'deletion': len('CACATG'),
        'insertion': 'C'
    })
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleB']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1984,
        'deletion': len('GTGATTG'),
        'insertion': 'TTGA'
    })
    assert 'OTHER' == v.var_type, 'Type is incorrect'
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleC']} == set(v.sample_ids)

    # Check to make sure unknown/missing positions are stored
    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 9,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None
    assert 'UNKNOWN_MISSING' == v.var_type
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleA'], sample_name_ids['SampleB'],
            sample_name_ids['SampleC']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 5100,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None
    assert 'UNKNOWN_MISSING' == v.var_type
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleA'], sample_name_ids['SampleC']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 873,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is None


def do_test_insert_variants_examine_variation_annotations(variation_service, sample_service,
                                                          database, snpeff_nucleotide_data_package):
    session = database.get_session()

    variation_service.insert(feature_scope_name='NC_011083', data_package=snpeff_nucleotide_data_package)

    sample_name_ids = sample_service.find_sample_name_ids({'SH10-014', 'SH14-001', 'SH14-014'})
    assert 3 == len(sample_name_ids.values())

    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'

    assert 177 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'

    # SNV/SNP
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'NC_011083',
        'position': 602110,
        'deletion': len('T'),
        'insertion': 'G'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SH10-014'], sample_name_ids['SH14-001'],
            sample_name_ids['SH14-014']} == set(v.sample_ids)

    assert 'missense_variant' == v.annotation
    assert 'MODERATE' == v.annotation_impact
    assert 'SEHA_RS03410' == v.annotation_gene_id
    assert 'SEHA_RS03410' == v.annotation_gene_name
    assert 'transcript' == v.annotation_feature_type
    assert 'protein_coding' == v.annotation_transcript_biotype
    assert 'c.1229T>G' == v.annotation_hgvs_c
    assert 'p.Leu410Arg' == v.annotation_hgvs_p
    assert 'hgvs:NC_011083:SEHA_RS03410:c.1229T>G' == v.id_hgvs_c
    assert 'hgvs:NC_011083:SEHA_RS03410:p.Leu410Arg' == v.id_hgvs_p

    # Deletion
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'NC_011083',
        'position': 833114,
        'deletion': len('GT'),
        'insertion': 'G'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SH14-001'],
            sample_name_ids['SH14-014']} == set(v.sample_ids)

    assert 'frameshift_variant' == v.annotation
    assert 'HIGH' == v.annotation_impact
    assert 'SEHA_RS04540' == v.annotation_gene_id
    assert 'glf' == v.annotation_gene_name
    assert 'transcript' == v.annotation_feature_type
    assert 'protein_coding' == v.annotation_transcript_biotype
    assert 'c.696delT' == v.annotation_hgvs_c
    assert 'p.Phe232fs' == v.annotation_hgvs_p
    assert 'hgvs:NC_011083:SEHA_RS04540:c.696delT' == v.id_hgvs_c
    assert 'hgvs:NC_011083:SEHA_RS04540:p.Phe232fs' == v.id_hgvs_p

    # Insertion
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'NC_011083',
        'position': 2380144,
        'deletion': len('A'),
        'insertion': 'AATTTTAT'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SH14-001'],
            sample_name_ids['SH14-014']} == set(v.sample_ids)

    assert 'frameshift_variant' == v.annotation
    assert 'HIGH' == v.annotation_impact
    assert 'SEHA_RS12315' == v.annotation_gene_id
    assert 'SEHA_RS12315' == v.annotation_gene_name
    assert 'transcript' == v.annotation_feature_type
    assert 'protein_coding' == v.annotation_transcript_biotype
    assert 'c.431_437dupTTTATAT' == v.annotation_hgvs_c
    assert 'p.Ter149fs' == v.annotation_hgvs_p
    assert 'hgvs:NC_011083:SEHA_RS12315:c.431_437dupTTTATAT' == v.id_hgvs_c
    assert 'hgvs:NC_011083:SEHA_RS12315:p.Ter149fs' == v.id_hgvs_p

    # MNP
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'NC_011083',
        'position': 3535656,
        'deletion': len('CTCAAAA'),
        'insertion': 'GTCATAG'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'MNP' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SH14-001'],
            sample_name_ids['SH10-014']} == set(v.sample_ids)

    assert 'missense_variant' == v.annotation
    assert 'MODERATE' == v.annotation_impact
    assert 'SEHA_RS17780' == v.annotation_gene_id
    assert 'oadA' == v.annotation_gene_name
    assert 'transcript' == v.annotation_feature_type
    assert 'protein_coding' == v.annotation_transcript_biotype
    assert 'c.582_588delTTTTGAGinsCTATGAC' == v.annotation_hgvs_c
    assert 'p.PheGlu195TyrAsp' == v.annotation_hgvs_p
    assert 'hgvs:NC_011083:SEHA_RS17780:c.582_588delTTTTGAGinsCTATGAC' == v.id_hgvs_c
    assert 'hgvs:NC_011083:SEHA_RS17780:p.PheGlu195TyrAsp' == v.id_hgvs_p


def test_insert_variants_examine_variation_annotations(database, snpeff_nucleotide_data_package,
                                                       reference_service_with_snpeff_data,
                                                       sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_snpeff_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    do_test_insert_variants_examine_variation_annotations(variation_service=variation_service,
                                                          sample_service=sample_service,
                                                          database=database,
                                                          snpeff_nucleotide_data_package=snpeff_nucleotide_data_package)


def test_insert_variants_examine_variation_annotations_parallel_variants(database,
                                                                         snpeff_nucleotide_data_package_parallel_variants,
                                                                         reference_service_with_snpeff_data,
                                                                         sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_snpeff_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    do_test_insert_variants_examine_variation_annotations(variation_service=variation_service,
                                                          sample_service=sample_service,
                                                          database=database,
                                                          snpeff_nucleotide_data_package=snpeff_nucleotide_data_package_parallel_variants)


def test_insert_variants_regular_vcf_reader_examine_variation(database, regular_nucleotide_data_package,
                                                              reference_service_with_data,
                                                              sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    session = database.get_session()

    variation_service.insert(feature_scope_name='genome', data_package=regular_nucleotide_data_package)

    sample_name_ids = sample_service.find_sample_name_ids({'SampleA', 'SampleB', 'SampleC'})
    assert 3 == len(sample_name_ids.values())

    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'

    assert 112 + 440 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'

    # Check to make sure some variants are stored
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1048,
        'deletion': len('C'),
        'insertion': 'G'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SampleA']} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1135,
        'deletion': len('CCT'),
        'insertion': 'C'
    })
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SampleB'], sample_name_ids['SampleC']} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 883,
        'deletion': len('CACATG'),
        'insertion': 'C'
    })
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SampleB']} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1984,
        'deletion': len('GTGATTG'),
        'insertion': 'TTGA'
    })
    assert 'OTHER' == v.var_type, 'Type is incorrect'
    assert {sample_name_ids['SampleC']} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    # Check to make sure unknown/missing positions are stored
    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 9,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None
    assert 'UNKNOWN_MISSING' == v.var_type
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleA'], sample_name_ids['SampleB'],
            sample_name_ids['SampleC']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 5100,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None
    assert 'UNKNOWN_MISSING' == v.var_type
    assert v.id_hgvs_c is None
    assert v.id_hgvs_p is None
    assert {sample_name_ids['SampleA'], sample_name_ids['SampleC']} == set(v.sample_ids)

    v = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 873,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is None


def test_insert_variants_regular_vcf_reader_examine_variation_no_unknown(database, regular_nucleotide_data_package,
                                                                         reference_service_with_data,
                                                                         sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)
    session = database.get_session()

    variation_service.insert(feature_scope_name='genome', data_package=regular_nucleotide_data_package)

    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'
    assert 112 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'


def test_insert_variants_duplicates(database, snippy_nucleotide_data_package, reference_service_with_data,
                                    sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)
    session = database.get_session()

    assert 0 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 0 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 0 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    assert 112 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 3 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'

    with pytest.raises(EntityExistsError) as execinfo:
        variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    assert 'Passed samples already have features for feature scope [genome]' in str(execinfo.value)
    assert 112 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 3 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'


def test_insert_variants_duplicates_subset(database, snippy_nucleotide_data_package, reference_service_with_data,
                                           sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)
    session = database.get_session()

    assert 0 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 0 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 0 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    assert 112 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 3 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'

    # Select a subset of samples
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        sample_dirs_subset = [data_dir / 'SampleA']
        subset_snippy_data_package = NucleotideSampleDataPackage.create_from_snippy(
            sample_dirs_subset,
            sample_files_processor=SerialSampleFilesProcessor(tmp_dir)
        )

        with pytest.raises(EntityExistsError) as execinfo:
            variation_service.insert(feature_scope_name='genome', data_package=subset_snippy_data_package)

        assert 'Passed samples already have features for feature scope [genome]' in str(execinfo.value)
        assert 112 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
        assert 3 == session.query(Sample).count(), 'Incorrect number of Samples'
        assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleSequences'


def test_multiple_insert_variants_different_samples(database, snippy_nucleotide_data_package_AB,
                                                    snippy_nucleotide_data_package_C,
                                                    reference_service_with_data,
                                                    sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=50)
    session = database.get_session()

    assert 0 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 0 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 0 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'

    # Insert samples A and B
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package_AB)

    assert 96 + 437 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 2 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 2 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'

    sampleA = session.query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = session.query(Sample).filter(Sample.name == 'SampleB').one()

    ## Variant which is present only in Sample A
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1708,
        'deletion': len('ATGCTGTTCAATAC'),
        'insertion': 'A'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sampleA.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Variant which is present only in Samples B
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 4539,
        'deletion': len('T'),
        'insertion': 'A'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Variant which is present only in Samples B and C
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 4585,
        'deletion': len('T'),
        'insertion': 'C'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample A
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 5071,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleA.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample B
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 4247,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample B
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 645,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    # Insert sample C
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package_C)

    assert 112 + 440 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 3 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 3 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'

    sampleC = session.query(Sample).filter(Sample.name == 'SampleC').one()

    ## Variant which is present only in Sample A
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1708,
        'deletion': len('ATGCTGTTCAATAC'),
        'insertion': 'A'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sampleA.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Variant which is present only in Samples B
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 4539,
        'deletion': len('T'),
        'insertion': 'A'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Variant which is present only in Samples B and C
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 4585,
        'deletion': len('T'),
        'insertion': 'C'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sampleB.id, sampleC.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample A
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 5071,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleA.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample B
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 4247,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample B
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 645,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleB.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    ## Missing only in Sample C
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1868,
        'deletion': 1,
        'insertion': '?'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'UNKNOWN_MISSING' == v.var_type, 'Type is incorrect'
    assert {sampleC.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])


def test_multiple_insert_variants_different_reference_genomes(database, snippy_nucleotide_data_package_AB,
                                                              snpeff_nucleotide_data_package,
                                                              reference_service_with_data,
                                                              sample_service, filesystem_storage):
    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=True,
                                         sql_select_limit=500)
    session = database.get_session()

    # Add snpeff reference genome
    reference_service_with_data.add_reference_genome(reference_file_snpeff)

    # Insert samples A and B, and snpeff samples
    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package_AB)
    variation_service.insert(feature_scope_name='NC_011083', data_package=snpeff_nucleotide_data_package)

    assert 96 + 437 + 177 == session.query(NucleotideVariantsSamples).count(), 'Incorrect number of storage entries'
    assert 5 == session.query(Sample).count(), 'Incorrect number of Samples'
    assert 5 == session.query(SampleNucleotideVariation).count(), 'Incorrect number of SampleNucleotideVariation'

    sampleA = session.query(Sample).filter(Sample.name == 'SampleA').one()
    sample_SH10_014 = session.query(Sample).filter(Sample.name == 'SH10-014').one()
    sample_SH14_001 = session.query(Sample).filter(Sample.name == 'SH14-001').one()
    sample_SH14_014 = session.query(Sample).filter(Sample.name == 'SH14-014').one()

    ## Variant which is present only in Sample A
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'reference',
        'position': 1708,
        'deletion': len('ATGCTGTTCAATAC'),
        'insertion': 'A'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'INDEL' == v.var_type, 'Type is incorrect'
    assert {sampleA.id} == set(v.sample_ids)
    assert all([x is None for x in [v.annotation, v.annotation_impact, v.annotation_gene_name, v.annotation_gene_id,
                                    v.annotation_feature_type, v.annotation_transcript_biotype, v.annotation_hgvs_c,
                                    v.annotation_hgvs_p]])

    # SNV/SNP for snpeff data
    v: NucleotideVariantsSamples = session.query(NucleotideVariantsSamples).get({
        'sequence': 'NC_011083',
        'position': 602110,
        'deletion': len('T'),
        'insertion': 'G'
    })
    assert v is not None, 'Particular storage does not exist'
    assert 'SNP' == v.var_type, 'Type is incorrect'
    assert {sample_SH10_014.id, sample_SH14_001.id, sample_SH14_014.id} == set(v.sample_ids)

    assert 'missense_variant' == v.annotation
    assert 'MODERATE' == v.annotation_impact
    assert 'SEHA_RS03410' == v.annotation_gene_id
    assert 'SEHA_RS03410' == v.annotation_gene_name
    assert 'transcript' == v.annotation_feature_type
    assert 'protein_coding' == v.annotation_transcript_biotype
    assert 'c.1229T>G' == v.annotation_hgvs_c
    assert 'p.Leu410Arg' == v.annotation_hgvs_p
    assert 'hgvs:NC_011083:SEHA_RS03410:c.1229T>G' == v.id_hgvs_c
    assert 'hgvs:NC_011083:SEHA_RS03410:p.Leu410Arg' == v.id_hgvs_p


def test_get_variants_ordered(database, snippy_nucleotide_data_package, reference_service_with_data,
                              sample_service, filesystem_storage):
    def find_variant_by_position(variants: List[NucleotideVariantsSamples], position: int) -> NucleotideVariantsSamples:
        for v in variants:
            if v.position == position:
                return v

        raise Exception(f'Could not find storage with position {position}')

    variation_service = VariationService(database_connection=database,
                                         reference_service=reference_service_with_data,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=500)

    variation_service.insert(feature_scope_name='genome', data_package=snippy_nucleotide_data_package)

    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    variants = variation_service.get_variants_ordered(sequence_name='reference')

    assert 60 == len(variants), 'Incorrect number of variants returned (counting only SNV/SNPs)'

    v1 = find_variant_by_position(variants, 4265)
    assert 'reference:4265:1:C' == v1.spdi, 'Incorrect storage returned'
    assert {sampleA.id} == set(v1.sample_ids)

    v2 = find_variant_by_position(variants, 839)
    assert 'reference:839:1:G' == v2.spdi, 'Incorrect storage returned'
    assert {sampleB.id, sampleC.id} == set(v2.sample_ids)


def test_get_sample_nucleotide_variation_one_sample(variation_service):
    sample_variations = variation_service.get_sample_nucleotide_variation(['SampleA'])

    assert 1 == len(sample_variations)
    assert isinstance(sample_variations[0], SampleNucleotideVariation)
    assert 'SampleA' == sample_variations[0].sample.name


def test_get_sample_nucleotide_variation_all_samples(variation_service):
    sample_variations = variation_service.get_sample_nucleotide_variation(['SampleA', 'SampleB', 'SampleC'])

    assert 3 == len(sample_variations)
    assert {'SampleA', 'SampleB', 'SampleC'} == {v.sample.name for v in sample_variations}


def test_count_mutations_in_sample_ids_one_sample(database, variation_service: VariationService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()

    expected_df = pd.read_csv(snippy_all_dataframes['SampleA'], sep='\t').groupby('Mutation').agg({
        'Sequence': 'first',
        'Position': 'first',
        'Deletion': 'first',
        'Insertion': 'first',
        'Mutation': 'count',
    }).rename(columns={'Mutation': 'Count'}).sort_index()
    print('Expected')
    print(expected_df)

    mutations_df = variation_service.count_mutations_in_sample_ids_dataframe([sampleA.id])
    mutations_df = mutations_df.sort_index()
    print('Actual')
    print(mutations_df)

    assert len(expected_df) == len(mutations_df)
    assert list(expected_df.columns) == list(mutations_df.columns)
    assert list(expected_df.index) == list(mutations_df.index)
    assert list(expected_df['Position']) == list(mutations_df['Position'])
    assert list(expected_df['Deletion']) == list(mutations_df['Deletion'])
    assert list(expected_df['Insertion']) == list(mutations_df['Insertion'])
    assert list(expected_df['Count']) == list(mutations_df['Count'])


def test_count_mutations_in_sample_ids_three_samples_only_snps(database, variation_service: VariationService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    dfA = pd.read_csv(snippy_snps_dataframes['SampleA'], sep='\t')
    dfB = pd.read_csv(snippy_snps_dataframes['SampleB'], sep='\t')
    dfC = pd.read_csv(snippy_snps_dataframes['SampleC'], sep='\t')
    expected_df = pd.concat([dfA, dfB, dfC])
    expected_df = expected_df.groupby('Mutation').agg({
        'Sequence': 'first',
        'Position': 'first',
        'Deletion': 'first',
        'Insertion': 'first',
        'Mutation': 'count',
    }).rename(columns={'Mutation': 'Count'}).sort_index()

    mutations_df = variation_service.count_mutations_in_sample_ids_dataframe([sampleA.id, sampleB.id, sampleC.id],
                                                                             mutation_type='snp')
    mutations_df = mutations_df.sort_index()

    assert len(expected_df) == len(mutations_df)
    assert list(expected_df.columns) == list(mutations_df.columns)
    assert list(expected_df.index) == list(mutations_df.index)
    assert list(expected_df['Position']) == list(mutations_df['Position'])
    assert list(expected_df['Deletion']) == list(mutations_df['Deletion'])
    assert list(expected_df['Insertion']) == list(mutations_df['Insertion'])
    assert list(expected_df['Count']) == list(mutations_df['Count'])


def test_count_mutations_in_sample_ids_three_samples_all_mutations(database, variation_service: VariationService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    dfA = pd.read_csv(snippy_all_dataframes['SampleA'], sep='\t')
    dfB = pd.read_csv(snippy_all_dataframes['SampleB'], sep='\t')
    dfC = pd.read_csv(snippy_all_dataframes['SampleC'], sep='\t')
    expected_df = pd.concat([dfA, dfB, dfC])
    expected_df = expected_df.groupby('Mutation').agg({
        'Sequence': 'first',
        'Position': 'first',
        'Deletion': 'first',
        'Insertion': 'first',
        'Mutation': 'count',
    }).rename(columns={'Mutation': 'Count'}).sort_index()

    mutations_df = variation_service.count_mutations_in_sample_ids_dataframe([sampleA.id, sampleB.id, sampleC.id],
                                                                             )
    mutations_df = mutations_df.sort_index()

    assert len(expected_df) == len(mutations_df)
    assert list(expected_df.columns) == list(mutations_df.columns)
    assert list(expected_df.index) == list(mutations_df.index)
    assert list(expected_df['Position']) == list(mutations_df['Position'])
    assert list(expected_df['Deletion']) == list(mutations_df['Deletion'])
    assert list(expected_df['Insertion']) == list(mutations_df['Insertion'])
    assert list(expected_df['Count']) == list(mutations_df['Count'])



import logging
from pathlib import Path
from typing import Optional

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.storage.util import TRACE_LEVEL

logger = logging.getLogger(__name__)


class NucleotideSampleDataSequenceMask(NucleotideSampleData):

    def __init__(self, sample_name: str, vcf_file: Path, vcf_file_index: Optional[Path],
                 sample_mask_sequence: Optional[Path], subtract_vcf_from_mask: bool = False):
        super().__init__(sample_name=sample_name,
                         vcf_file=vcf_file,
                         vcf_file_index=vcf_file_index,
                         mask_bed_file=None,
                         preprocessed=False)
        self._sample_mask_sequence = sample_mask_sequence
        self._subtract_vcf_from_mask = subtract_vcf_from_mask

    def _preprocess_mask(self, output_dir: Path) -> Path:
        if self._sample_mask_sequence is None:
            mask_file = output_dir / f'{self.sample_name_persistence}.bed.gz'
            mask = MaskedGenomicRegions.empty_mask()
            mask.write(mask_file)
            return mask_file
        else:
            new_file = output_dir / f'{self.sample_name_persistence}.bed.gz'
            if new_file.exists():
                raise Exception(f'File {new_file} already exists')

            name, records = SequenceFile(self._sample_mask_sequence).parse_sequence_file()
            logger.log(TRACE_LEVEL, f'Getting genomic masks from {self._sample_mask_sequence}')
            masked_regions = MaskedGenomicRegions.from_sequences(sequences=records)

            if self._subtract_vcf_from_mask:
                logger.log(TRACE_LEVEL, f'Subtracting variants in vcf_file=[{self._vcf_file}] '
                                        f'from mask produced by sequence=[{self._sample_mask_sequence}]')
                vcf_regions = MaskedGenomicRegions.from_vcf_file(self._vcf_file)
                masked_regions = masked_regions.subtract(vcf_regions)

            masked_regions.write(new_file)
            return new_file

    @classmethod
    def create(cls, sample_name: str, vcf_file: Path,
               sample_mask_sequence: Optional[Path] = None,
               subtract_vcf_from_mask: bool = False) -> NucleotideSampleData:
        return NucleotideSampleDataSequenceMask(sample_name=sample_name,
                                                vcf_file=vcf_file,
                                                vcf_file_index=None,
                                                sample_mask_sequence=sample_mask_sequence,
                                                subtract_vcf_from_mask=subtract_vcf_from_mask)



from pathlib import Path

from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile

# Reads and prepares a reference genome by writing it to a FASTA file.

input_reference = Path(snakemake.input.reference)
output_reference = Path(snakemake.output.reference)

sequence_file = SequenceFile(input_reference)
sequence_file.write(output_reference, 'fasta')



from pathlib import Path

from genomics_data_index.storage.io.mutation.SnpEffDatabase import SnpEffDatabase

# Annotate VCF with snpeff

snpeff_database_conf = Path(snakemake.input.snpeff_db_conf)
snpeff_database = SnpEffDatabase.create_from_config(snpeff_database_conf)
input_vcf = Path(snakemake.input.vcf).absolute()
output_vcf = Path(snakemake.output.snpeff_vcf).absolute()

snpeff_database.annotate(input_vcf_file=input_vcf, output_vcf_file=output_vcf)



import os


# Generates a file of file names (fofn) from the given samples and VCF/consensus files
# This file of file names is then used for input to the genomic indexer (gdi)


# From https://stackoverflow.com/a/51779415
def remove_suffix(s, suffix):
    return s[:-len(suffix)] if s.endswith(suffix) else s


# TODO: Instead of parallel lists I should probably have a dictionary keyed by the sample name
sample_vcfs = snakemake.input.sample_vcfs
sample_masks = snakemake.input.sample_masks
sample_sketches = snakemake.input.sample_sketches
include_kmer = snakemake.config["include_kmer"]
output_file = snakemake.output[0]

if len(sample_vcfs) != len(sample_masks):
    raise Exception(f'sample_vcfs length [{len(sample_vcfs)}] '
                    f'does not equal sample_masks length [{len(sample_masks)}]')

if include_kmer and len(sample_vcfs) != len(sample_sketches):
    raise Exception(f'sample_vcfs length [{len(sample_vcfs)}] '
                    f'does not equal sample_sketches length [{len(sample_sketches)}]')

with open(output_file, 'w') as fh:
    if include_kmer:
        fh.write('Sample\tVCF\tMask File\tSketch File\n')
    else:
        fh.write('Sample\tVCF\tMask File\n')

    for idx in range(0, len(sample_masks)):
        vcf = os.path.abspath(sample_vcfs[idx])
        mask = os.path.abspath(sample_masks[idx])

        # I check both sample names to guarantee that files don't get mixed up
        # I don't know enough about snakemake to solve this differently. Ideally I'd like to just pass a dictionary
        # to this script which maps the sample name to the vcf and consensus file
        sample_name_vcf = remove_suffix(os.path.basename(vcf), '.vcf.gz')
        sample_name_mask = remove_suffix(os.path.basename(mask), '.bed.gz')

        if sample_name_mask != sample_name_vcf:
            raise Exception(f'VCF=[{vcf}] has a different sample name from mask=[{mask}]')

        if include_kmer:
            sketch = os.path.abspath(sample_sketches[idx])
            sample_name_sketch = remove_suffix(os.path.basename(sketch), '.sig.gz')

            if sample_name_vcf != sample_name_sketch:
                raise Exception(f'VCF=[{vcf}] has a different sample name from sketch=[{sketch}]')

            fh.write(f'{sample_name_vcf}\t{vcf}\t{mask}\t{sketch}\n')
        else:
            fh.write(f'{sample_name_vcf}\t{vcf}\t{mask}\n')



import math

import pytest

from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.db import Sample
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.KmerService import KmerService


def test_find_matches_all(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=31,
                                                                 distance_threshold=1.0)

    assert {sampleA.id, sampleB.id, sampleC.id} == set(all_matches_set)


def test_find_matches_all_in_sub_universe(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    samples_universe = SampleSet([sampleB.id, sampleC.id])

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=31,
                                                                 distance_threshold=1.0,
                                                                 samples_universe=samples_universe)

    assert {sampleB.id, sampleC.id} == set(all_matches_set)


def test_find_matches_all_in_empty_universe(database: DatabaseConnection, kmer_service_with_data: KmerService):
    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=31,
                                                                 distance_threshold=1.0,
                                                                 samples_universe=SampleSet.create_empty())
    assert set() == set(all_matches_set)


def test_find_matches_all_lower_threshold(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=31,
                                                                 distance_threshold=0.522)

    assert {sampleA.id, sampleB.id, sampleC.id} == set(all_matches_set)


def test_find_matches_two(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=31,
                                                                 distance_threshold=0.521)

    assert {sampleA.id, sampleC.id} == set(all_matches_set)


def test_find_matches_one(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=31,
                                                                 distance_threshold=0.49)

    assert {sampleA.id} == set(all_matches_set)


def test_find_matches_all_lower_different_genome(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleC'], kmer_size=31,
                                                                 distance_threshold=0.5)

    assert {sampleA.id, sampleB.id, sampleC.id} == set(all_matches_set)


def test_find_matches_one_different_genome(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleC'], kmer_size=31,
                                                                 distance_threshold=0.49)

    assert {sampleB.id, sampleC.id} == set(all_matches_set)


def test_find_matches_two_different_genomes(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA', 'SampleC'], kmer_size=31,
                                                                 distance_threshold=0.49)

    assert {sampleA.id, sampleB.id, sampleC.id} == set(all_matches_set)


def test_find_matches_different_kmer(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    all_matches_set = kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=21,
                                                                 distance_threshold=0.4)

    assert {sampleA.id, sampleC.id} == set(all_matches_set)


def test_find_matches_nonindexed_kmer_size(database: DatabaseConnection, kmer_service_with_data: KmerService):
    with pytest.raises(Exception) as execinfo:
        kmer_service_with_data.find_matches_within(['SampleA'], kmer_size=11,
                                                   distance_threshold=1.0)
    assert 'Could not run' in str(execinfo.value)


def test_distance_matrix_all(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    sample_set = SampleSet([sampleA.id, sampleB.id, sampleC.id])

    results_d, labels = kmer_service_with_data.get_distance_matrix(kmer_size=31,
                                                                   sample_ids=sample_set)

    assert (3, 3) == results_d.shape
    assert ['SampleA', 'SampleB', 'SampleC'] == labels

    l = {element: idx for idx, element in enumerate(labels)}

    assert math.isclose(results_d[l['SampleA']][l['SampleA']], 0, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleA']][l['SampleB']], 0.522, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleA']][l['SampleC']], 0.5, rel_tol=1e-3)

    assert math.isclose(results_d[l['SampleB']][l['SampleA']], 0.522, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleB']][l['SampleB']], 0, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleB']][l['SampleC']], 0.3186, rel_tol=1e-3)

    assert math.isclose(results_d[l['SampleC']][l['SampleA']], 0.5, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleC']][l['SampleB']], 0.3186, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleC']][l['SampleC']], 0, rel_tol=1e-3)


def test_distance_matrix_two(database: DatabaseConnection, kmer_service_with_data: KmerService):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()

    sample_set = SampleSet([sampleA.id, sampleB.id])

    results_d, labels = kmer_service_with_data.get_distance_matrix(kmer_size=31,
                                                                   sample_ids=sample_set)

    assert (2, 2) == results_d.shape
    assert ['SampleA', 'SampleB'] == labels

    l = {element: idx for idx, element in enumerate(labels)}

    assert math.isclose(results_d[l['SampleA']][l['SampleA']], 0, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleA']][l['SampleB']], 0.522, rel_tol=1e-3)

    assert math.isclose(results_d[l['SampleB']][l['SampleA']], 0.522, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleB']][l['SampleB']], 0, rel_tol=1e-3)
